"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Class for family type data storage class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Used as storage class when retrieving family type data from a family file.

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Revit.Family.Data.Objects import ifamily_data_storage as IFamDataStorage


class FamilyTypeDataStorage(IFamDataStorage.IFamilyDataStorage):

    # data type for this class ( used in reports as first entry per row )
    data_type = "FamilyType"

    # number of properties in this class ( used in report reader function )
    number_of_properties = 8

    def __init__(
        self,
        root_name_path,  # root name path ( most often the same as the family name)
        root_category_path,  # family category
        family_name,  # family name
        family_file_path,  # family file path
        family_type_name,  # family type name
        parameters,  # parameters and there values for this type
        last_updated_date=None,  # last updated date (when was the data retrieved from the family type)
        last_updated_time=None,  # last updated time (when was the data retrieved from the family type)
        **kwargs
    ):
        """
        constructor

        :param root_name_path: root name path ( most often the same as the family name)
        :param root_category_path: family category
        :param family_name: family name
        :param family_file_path: family file path
        :param family_type_name: family type name
        :param parameters: parameters and there values for this type
        :param last_updated_date: last updated date (when was the data retrieved from the family type)
        :param last_updated_time: last updated time (when was the data retrieved from the family type)
        """

        # store args in base class
        super(FamilyTypeDataStorage, self).__init__(
            # data_type=FamilyTypeDataStorage.data_type,
            data_type=FamilyTypeDataStorage.data_type,
            root_name_path=root_name_path,
            root_category_path=root_category_path,
            family_name=family_name,
            family_file_path=family_file_path,
        )

        self._family_type_name = family_type_name
        self._parameters = parameters
        self._last_updated_date = last_updated_date  # last updated date (when was the data retrieved from the family type)
        self._last_updated_time = last_updated_time  # last updated time (when was the data retrieved from the family type)

    @property
    def family_type_name(self):
        return self._family_type_name

    @property
    def parameters(self):
        return self._parameters

    @property
    def last_updated_date(self):
        return self._last_updated_date

    @property
    def last_updated_time(self):
        return self._last_updated_time

    def get_parameter_by_name(self, parameter_name):
        """
        Returns the parameter by its name.

        :param parameter_name: The name of the parameter to retrieve.
        :return: The parameter if found, otherwise None.
        """
        for parameter in self._parameters:
            if parameter.name == parameter_name:
                return parameter
        return None

    def __eq__(self, other):
        """
        equal compare (ignores last updated date and time)

        :param other: object to compare with
        :return: True if equal, False if not
        """

        if not isinstance(other, FamilyTypeDataStorage):
            return NotImplemented
        return (
            self.data_type == other.data_type,
            self.root_name_path == other.root_name_path,
            self.root_category_path == other.root_category_path,
            self.family_name == other.family_name,
            self.family_file_path == other.family_file_path,
            self.family_type_name == other.family_type_name,
            self.parameters == other.parameters,
        )

    def __ne__(self, other):
        """
        not equal compare (ignores last updated date and time)

        :param other: object to compare with
        :return: True if not equal, False if equal
        """
        return not self.__eq__(other)

    def __hash__(self):
        """
        hash function for this class (ignores last updated date and time)

        :return: hash value
        """

        return hash(
            (
                self.data_type,
                self.root_name_path,
                self.root_category_path,
                self.family_name,
                self.family_file_path,
                self.family_type_name,
                self.parameters,
            )
        )

    def is_match_by_names_and_category(self, other):
        """
        check if this object is a match with another object by family name and category
        ignores:

        - last updated date and time
        - parameters
        - family file path

        :param other: object to compare with
        :return: True if match, False if not
        """

        if not isinstance(other, FamilyTypeDataStorage):
            return False

        return (
            self.root_name_path == other.root_name_path
            and self.root_category_path == other.root_category_path
            and self.family_name == other.family_name
            and self.family_type_name == other.family_type_name
        )

    def get_comparison_report_parameter_values(self, other, ignore_property_names=[]):
        """
        get the difference between this object and another family type data storage object in terms of parameter values only
        Assumes that the objects are equal in terms of family name, family type name, root name path and root category path

        :param other: object to compare with
        :return: list of properties that are different
        """

        if not isinstance(other, FamilyTypeDataStorage):
            return NotImplemented

        return_list = []

        # repeats per parameter difference
        diff_base = [
            self.family_name,
            self.root_category_path,
            "Found match in library",
            self.family_type_name,
            "Found match in library",
        ]

        for param in self.parameters:
            param_other = other.get_parameter_by_name(param.name)
            if param_other is None:
                # parameter not found in other object
                diff_parameter = [param.name, "No match in library"]
                return_list.append(diff_base + diff_parameter)
            else:
                if param != param_other and param.name not in ignore_property_names:
                    para_diff = param.get_difference(param_other)
                    for single_difference in para_diff:
                        # write each difference as a separate entry
                        parameter_diff_entry = [
                            param.name,
                            "Found match in library",
                            single_difference,
                        ]
                        return_list.append(diff_base + parameter_diff_entry)

        return return_list

    def get_difference(self, other, ignore_property_names=[]):
        """
        get the difference between this object and another family type data storage object

        :param other: object to compare with
        :return: list of properties that are different
        """

        if not isinstance(other, FamilyTypeDataStorage):
            return NotImplemented

        # holds all the differences per para
        diff = []

        # should return a list with an entry for each difference
        # if no differences are found the list should be empty

        if (
            self.data_type != other.data_type
            and "data_type" not in ignore_property_names
        ):
            diff.append("data_type: {} != {}".format(self.data_type, other.data_type))

        if (
            self.root_name_path != other.root_name_path
            and "root_name_path" not in ignore_property_names
        ):
            diff.append(
                "root_name_path: {} != {}".format(
                    self.root_name_path, other.root_name_path
                )
            )

        if (
            self.root_category_path != other.root_category_path
            and "root_category_path" not in ignore_property_names
        ):
            diff.append(
                "root_category_path: {} != {}".format(
                    self.root_category_path, other.root_category_path
                )
            )

        if (
            self.family_name != other.family_name
            and "family_name" not in ignore_property_names
        ):
            diff.append(
                "family_name: {} != {}".format(self.family_name, other.family_name)
            )

        if (
            self.family_file_path != other.family_file_path
            and "family_file_path" not in ignore_property_names
        ):
            diff.append(
                "family_file_path: {} != {}".format(
                    self.family_file_path, other.family_file_path
                )
            )

        if (
            self.family_type_name != other.family_type_name
            and "family_type_name" not in ignore_property_names
        ):
            diff.append(
                "family_type_name: {} != {}".format(
                    self.family_type_name, other.family_type_name
                )
            )

        # compare parameters by their names
        # if a match is found compare the values

        for param in self.parameters:
            param_other = other.get_parameter_by_name(param.name)
            if param_other is None:
                diff.append("parameter {} not found in other object".format(param.name))
            else:
                if param != param_other and param.name not in ignore_property_names:
                    para_diff = param.get_difference(param_other)
                    for single_difference in para_diff:
                        diff.append(
                            "parameter {} is different: {}".format(
                                param.name, single_difference
                            )
                        )
                else:
                    pass
                    # diff.append("parameter {} is equal. {}".format(param.name, para_diff))

        return diff


    def get_report_data(self, file_name=None):
        """
        get the data for the report

        :param file_name: name of the file the data was retrieved from
        :type file_name: str

        :return: list of data for the report
        """

        data = []
        
        for param in self.parameters:
            # get the report data for the parameter
            para_report_data = param.get_report_data()
            
            # add the file name if available
            if file_name is None:
                # if no file name is provided, use the family file path as this part atom most likely was created from a family file and not a project file
                file_name = self.family_file_path if self.family_file_path else "N/A"
            
            # build the default type information repeated for each parameter
            def_report_data = [
                file_name,
                self.family_name,
                self.root_category_path,
                self.family_type_name,
                self.last_updated_date,
                self.last_updated_time,
            ]

            index = len(def_report_data)
            # add the parameter report data to the default type information
            def_report_data[index:index] = para_report_data

            # append the data to the list
            data.append(def_report_data)
        return data
    

    def get_catalogue_file_data(self, parameter_names=None):
        """
        Get the data for the catalogue file

        :param parameter_names: list of parameter names to include in the catalogue file and their order
        :type parameter_names: [str]
        :return: list of data for the catalogue file
        """

        data = []
        
        if parameter_names is None:
            for param in self.parameters:
                # get the report data for the parameter
                para_report_data = param.get_catalogue_file_data()
                # append the data to the list
                data.append(para_report_data)
        else:
            for parameter_name in parameter_names:
                param = self.get_parameter_by_name(parameter_name)
                if param is not None:
                    # get the report data for the parameter
                    para_report_data = param.get_catalogue_file_data()
                    # append the data to the list
                    data.append(para_report_data)
                else:
                    # add no data for the parameter
                    pass
                
        # add the type name to the front
        data.insert(0, self.family_type_name)

        return data
    
    def get_catalogue_file_header_row(self, parameter_names):
        """
        Get the header row for the catalogue file

        :param parameter_names: list of parameter names to include in the catalogue file and their order
        :type parameter_names: [str]
        :return: list of data for the catalogue file header row
        """

        # set up a header row with empty entry for the first column
        header_row = [""]

        # add the header row based on the parameter names
        if parameter_names is not None:
            for parameter_name in parameter_names:

                para_storage = self.get_parameter_by_name(parameter_name)
                if para_storage is not None:
                    # get the header row for the parameter
                    para_header_row = para_storage.get_catalogue_file_header_row()
                    # append the data to the list
                    header_row.append(para_header_row)
                else:
                    # add no data for the parameter
                    pass
        else:
            for param in self.parameters:
                # get the header row for the parameter
                para_header_row = param.get_catalogue_file_header_row()
                # append the data to the list
                header_row.append(para_header_row)

        return header_row


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family line pattern data processor class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Revit.Family.Data.Objects.ifamily_processor import IFamilyProcessor
from duHast.Revit.LinePattern.Data.Objects import line_pattern_data as rLinePatData
from duHast.Revit.LinePattern.Data.Objects.line_pattern_data_processor_defaults import (
    DATA_TYPE_PROCESSOR as data_type_line_pattern_processor,
)
from duHast.Revit.Family.Data.Objects import ifamily_data as IFamData
from duHast.Revit.Family.Data.Objects.ifamily_data_storage import IFamilyDataStorage
from duHast.Utilities.Objects import result as res


class LinePatternProcessor(IFamilyProcessor):

    data_type = data_type_line_pattern_processor

    def __init__(self, pre_actions=None, post_actions=None):
        """
        Class constructor.
        """

        # store data type  in base class
        super(LinePatternProcessor, self).__init__(
            pre_actions=pre_actions,
            post_actions=[self._post_action_update_used_line_patterns],
            data_type=LinePatternProcessor.data_type,
        )

        # set default post action to updated line patterns used in root processor with any line patterns found in nested
        # families
        # self.postActions = [self._postActionUpdateUsedLinePatterns]
        # add any other post actions
        if post_actions != None:
            for post_action in post_actions:
                self.post_actions.append(post_action)

    def process(self, doc, root_path, root_category_path):
        """
        Calls processor instance with the document and root path provided and adds processor instance to class property .data

        :param doc: Current family document.
        :type doc: Autodesk.Revit.DB.Document
        :param rootPath: The path of the nested family in a tree: rootFamilyName::nestedFamilyNameOne::nestedFamilyTwo\
            This includes the actual family name as the last node.
        :type rootPath: str
        :param rootCategoryPath: The categroy path of the nested family in a tree: rootFamilyCategory::nestedFamilyOneCategory::nestedFamilyTwoCategory\
            This includes the actual family category as the last node.
        :type rootCategoryPath: str
        """

        dummy = rLinePatData.LinePatternData(root_path, root_category_path)
        dummy.process(doc)
        self.data.append(dummy)

    # TODO: this could go into a base class function
    def _is_sub_line_pattern_present(
        self, root_family_data, nested_family_line_pattern_storage
    ):
        """
        Check if a line pattern is already present in the root family data.

        :param root_family_data: The root family data.
        :type root_family_data: list
        :param nested_family_line_pattern_storage: The nested family line pattern data storage.
        :type nested_family_line_pattern_storage: IFamilyDataStorage
        :return: The root family data if the line pattern is present, otherwise None.
        :rtype: IFamData.IFamilyData
        """

        # check what came is a list
        if isinstance(root_family_data, list) == False:
            raise ValueError(
                "Root family data must be a list but is type: {}".format(
                    type(root_family_data)
                )
            )

        # check what came is a IFamilyDataStorage instance
        # I could check for the exact data storage type here(?)
        if isinstance(nested_family_line_pattern_storage, IFamilyDataStorage) == False:
            raise ValueError(
                "Nested family line pattern storage must be an instance of IFamilyDataStorage but is type: {}".format(
                    type(nested_family_line_pattern_storage)
                )
            )

        match = None
        for root_fam in root_family_data:
            if isinstance(root_fam, IFamData.IFamilyData) == False:
                raise ValueError(
                    "Root family data must be a list of IFamilyData objects but is type: {}".format(
                        type(root_fam)
                    )
                )
            storage_instances = root_fam.get_data()
            for storage_root in storage_instances:
                if (
                    storage_root.pattern_name
                    == nested_family_line_pattern_storage.pattern_name
                ):
                    match = root_fam
                    break
        return match

    # TODO: this could go into a base class function
    def _update_root_family_data(self, root_family_data, nested_families_line_patterns):
        """
        Update the root family data with the nested family line patterns.

        :param root_family_data: The root family data.
        :type root_family_data: list
        :param nested_families_line_patterns: The nested families line patterns.
        :type nested_families_line_patterns: list
        """

        # check what came is a list
        if isinstance(root_family_data, list) == False:
            raise ValueError(
                "Root family data must be a list but is type: {}".format(
                    type(root_family_data)
                )
            )
        if isinstance(nested_families_line_patterns, list) == False:
            raise ValueError(
                "Nested families line patterns must be a list but is type: {}".format(
                    type(nested_families_line_patterns)
                )
            )

        # loop over nested family line pattern data
        for nested_line_pattern_storage in nested_families_line_patterns:
            # check this is an instance of IFamilyDataStorage
            if isinstance(nested_line_pattern_storage, IFamilyDataStorage) == False:
                raise ValueError(
                    "Nested family line pattern must be an instance of IFamilyDataStorage but is type: {}".format(
                        type(nested_line_pattern_storage)
                    )
                )

            # check if pattern is already in root family
            matching_root_fam_pattern_data = self._is_sub_line_pattern_present(
                root_family_data, nested_line_pattern_storage
            )
            if matching_root_fam_pattern_data != None:
                root_storage_all = matching_root_fam_pattern_data.get_root_storage()
                # some data instances might have more than one root storage instance to represent multiple shared parameters present in the family
                for root_storage in root_storage_all:
                    # update used by list
                    if (
                        nested_line_pattern_storage.pattern_name
                        == root_storage.pattern_name
                    ):
                        # add the root path to the used by list for ease of identification of the origin of this pattern usage
                        root_storage.update_usage(nested_line_pattern_storage)
            else:
                pass
                # nothing to do if that pattern has not been reported to start off with
                # this patter could, for example, belong to the section marker family present in most 3d families

    # TODO: this could go into a base class function
    def _get_used_line_patterns(self, data_instances):
        """
        Get all used line patterns from the data instances.

        :param data_instances: List of IFamilyData objects.
        :type data_instances: list

        :return: List of storage instances representing used line patterns.
        :rtype: list
        """

        # data_instances should be a list of IFamilyData objects
        if isinstance(data_instances, list) == False:
            raise ValueError(
                "Data must be a list but is type: {}".format(type(data_instances))
            )

        used_items = []
        for d in data_instances:
            # check what is being processed...should be an IFamilyData object
            if isinstance(d, IFamData.IFamilyData) == False:
                raise ValueError(
                    "Data must be an instance of IFamilyData but is type: {}".format(
                        type(d)
                    )
                )

            # loop over storage of the data instance
            data_storage_instances = d.get_data()
            for storage in data_storage_instances:
                if storage.use_counter > 0:
                    used_items.append(storage)
        return used_items

    def _post_action_update_used_line_patterns(self, doc):
        return_value = res.Result()
        try:
            # find all line patterns of nested families
            nested_family_data = self._find_nested_families_data()
            # get used sub categories from nested data
            nested_family_used_line_patterns = self._get_used_line_patterns(
                nested_family_data
            )
            # update root family data only
            rootFamilyData = self._find_root_family_data()
            # update root processor data as required
            self._update_root_family_data(
                rootFamilyData, nested_family_used_line_patterns
            )
            return_value.update_sep(
                True, "Post Action Update line pattern data successful completed."
            )
        except Exception as e:
            return_value.update_sep(
                False,
                "Post Action Update line pattern data failed with exception: " + str(e),
            )
        return return_value



# import pkgutil

# __all__ = []
# for loader, module_name, is_pkg in pkgutil.walk_packages(__path__):
#    __all__.append(module_name)
#    _module = loader.find_module(module_name).load_module(module_name)
#    globals()[module_name] = _module



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Interface for family data storage used by class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Notes:

used to store used by other nested families data in root family data storage class.


"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects import base


class IFamilyDataStorageUsedBy(base.Base):
    def __init__(self, data_type, family_name, element_id, j, **kwargs):

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(IFamilyDataStorageUsedBy, self).__init__(**kwargs)

        if isinstance(data_type, str):
            self.data_type = data_type
        else:
            raise ValueError("data_type must be a string")

        if isinstance(family_name, str):
            self.root_name_path = family_name
        else:
            raise ValueError("family_name must be a string")

        if isinstance(element_id, int):
            self.element_id = element_id
        else:
            raise ValueError(
                "element_id [{}] must be an int but is {}".format(
                    element_id, type(element_id)
                )
            )

        # ini using JSON data
        if isinstance(j, dict):
            try:
                self.root_name_path = j.get("root_name_path", self.root_name_path)
                self.element_id = j.get("element_id", self.element_id)
                self.data_type = j.get("data_type", self.data_type)
            except Exception as e:
                print("Failed to initialise object with JSON data: {}".format(e))
                raise type(e)("Node {} failed to initialise with: {}".format(self.data_type, e))
    def __eq__(self, other):
        """
        Custom compare is equal override.

        :param other: Another instance of IFamilyDataStorageUsedBy base class
        :type other: :class:`.IFamilyDataStorageUsedBy`
        :return: True if all properties of compared class instances are equal, otherwise False.
        :rtype: Bool
        """

        return isinstance(other, IFamilyDataStorageUsedBy) and (
            self.data_type,
            self.root_name_path,
            self.element_id,
        ) == (
            other.data_type,
            other.root_name_path,
            other.element_id,
        )

    # python 2.7 needs custom implementation of not equal
    def __ne__(self, other):
        return not self.__eq__(other=other)



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A base class used to store view filters.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Stores view filters ( ParameterFilterElement in Revit)



ParameterFilterElement contains: 
    - ElementFilters ( a base class for filters)
    those in turn contain ElementLogicalFilter ( a base class for ) logical filters There are two types:
        - LogicalAndFilter
        - LogicalOrFilter
        and those in turn contain ElementParameterFilter ( a base class for parameter filters)
            - most often in view filters we seem to use ElementParameterFilter which consists of list of  FilterRule


"""


#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


import json

from duHast.Utilities.Objects import base
from duHast.Revit.Views.Objects.Data.view_filter_logic_container import ViewFilterLogicContainer

class ViewFilter(base.Base):
    def __init__(self, data_type="view filter", j=None, **kwargs):
        """
        Class constructor.

        """

        super(ViewFilter, self).__init__(**kwargs)

        self.data_type = data_type
        self.category_ids = []  # list of category ids the filter applies to
        self.logic_container = None  # should just be one
        self.name = ""
       

        # check if any data was past in with constructor!
        if j is not None:
            # check type of data that came in:
            if type(j) == str:
                # a string
                j = json.loads(j)
            elif type(j) == dict:
                # no action required
                pass
            else:
                raise ValueError(
                    "Argument supplied must be of type string or type dictionary"
                )

            # load values and throw exception if something is missing!
            try:
                # get category ids
                self.category_ids = j["category_ids"]

                # get containers
                self.logic_container = ViewFilterLogicContainer(j["logic_containers"])
               
            except Exception as e:
                raise ValueError(
                    "Node {} failed to initialise with: {}".format(
                        "OverrideByBase.data_type", e
                    )
                )


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around Revit detail items.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr

clr.AddReference("System.Core")
from System import Linq

clr.ImportExtensions(Linq)

# import common library modules
from duHast.Revit.Common import parameter_get_utils as rParaGet
from duHast.Revit.DetailItems.Utility import (
    detail_items_type_sorting as rDetailItemTypeSort,
)

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInCategory,
    BuiltInParameter,
    ElementId,
    FilteredElementCollector,
)

#: class name Autodesk.Revit.DB.ElementType
ELEMENT_TYPE = "Autodesk.Revit.DB.ElementType"
#: class name Autodesk.Revit.DB.FilledRegionType
FILLED_REGION_TYPE = "Autodesk.Revit.DB.FilledRegionType"
#: class name Autodesk.Revit.DB.FamilySymbol
FAMILY_SYMBOL = "Autodesk.Revit.DB.FamilySymbol"

#: List of class names which can be detailed components
DETAIL_COMPONENT_TYPES = [ELEMENT_TYPE, FILLED_REGION_TYPE, FAMILY_SYMBOL]



# -------------------------------- detail components -------------------------------------------------------


def get_all_detail_types_by_category(doc):
    """
    Gets all detail component types in the model.

    Filters by built in category.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A filtered element collector containing detail component types.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = (
        FilteredElementCollector(doc)
        .OfCategory(BuiltInCategory.OST_DetailComponents)
        .WhereElementIsElementType()
    )
    return collector


# -------------------------------- repeating detail types -------------------------------------------------------


def get_all_repeating_detail_type_ids_available(doc):
    """
    Get all repeating detail type id's in model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of element ids representing repeating detail types.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    dic = rDetailItemTypeSort.build_detail_type_ids_dictionary(
        get_all_detail_types_by_category(doc)
    )
    if ELEMENT_TYPE in dic:
        return dic[ELEMENT_TYPE]
    else:
        return []


# -------------------------------- Detail families -------------------------------------------------------


def get_all_detail_symbol_ids_available(doc):
    """
    Gets all detail symbol (types) ids in model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of element ids representing detail symbols.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    dic = rDetailItemTypeSort.build_detail_type_ids_dictionary(
        get_all_detail_types_by_category(doc)
    )
    if FAMILY_SYMBOL in dic:
        return dic[FAMILY_SYMBOL]
    else:
        return []


def get_detail_symbols_used_in_repeating_details(doc, ids_repeat_det):
    """
    Gets the ids of all symbols used in repeating details.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param idsRepeatDet: List of repeating detail type ids.
    :type idsRepeatDet: list Autodesk.Revit.DB.ElementIds

    :return: List of family symbol (type) ids.
    :rtype: list Autodesk.Revit.DB.ElementIds
    """

    ids = []
    for id_r in ids_repeat_det:
        repeat_detail = doc.GetElement(id_r)
        id = rParaGet.get_built_in_parameter_value(
            repeat_detail, BuiltInParameter.REPEATING_DETAIL_ELEMENT
        )
        if id not in ids and id != ElementId.InvalidElementId and id != None:
            ids.append(id)
    return ids



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions relating to combining text files. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import glob
import os
import csv

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import get_file_name_without_ext
from duHast.Utilities.files_get import get_files_single_directory
from duHast.Utilities.files_base_read import (
    read_column_based_text_file as read_column_based_text_file_base,
)
from duHast.Utilities.files_base_read import (
    get_unique_headers as get_unique_headers_base,
)
from duHast.Utilities.files_base_write import (
    write_report_data as write_report_data_base,
)


def combine_files(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    output_file_name="result.txt",
    file_getter=get_files_single_directory,
    delimiter=",",
    quoting=csv.QUOTE_MINIMAL,
):
    """
    Combines multiple text files into a single new file.
    Assumes:

    - all files have a header row followed by data rows
    - same number of headers (columns) in each files.
    - files have the same header names per column
    - files are encoded in UTF-8!

    The new file will be saved into the same folder as the original files.

    :param folder_path: Folder path from which to get files to be combined and to which the combined file will be saved.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :param output_file_name: The file name of the combined file, defaults to 'result.txt'
    :type output_file_name: str, optional
    :param file_getter: Function returning list of files to be combined, defaults to GetFilesSingleFolder
    :type file_getter: func(folder_path, file_prefix, file_suffix, file_extension), optional
    :param delimiter: The delimiter used in the files (e.g., ',' for CSV, '\t' for tab-separated), defaults to ','
    :type delimiter: str, optional
    :param quoting: The quoting option for the CSV writer, defaults to csv.QUOTE_MINIMAL
    :type quoting: int, optional

    :return:
        Result class instance.

        - result.status (bool) True if file content was combined without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will be an empty list.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()
    try:
        # check a file getter function was provided
        if file_getter is None:
            return_value.update_sep(False, "No file getter function provided.")
            return return_value

        # get files to combine using file getter function
        file_list = file_getter(folder_path, file_prefix, file_suffix, file_extension)

        # loop over file and combine...
        for file_index, file_ in enumerate(file_list):
            try:
                #line_counter = 0
                
                # attempt to read the file
                lines_result = read_column_based_text_file_base(file_, delimiter=delimiter)
                if lines_result.status is False:
                    return_value.update_sep(False, "Failed to read file: {} with {}".format(file_, lines_result.message))
                    # skip to next file
                    continue
                
                # get the lines read from the file
                lines = lines_result.result
                
                # determine write mode, default is append
                write_mode = "a"
                if file_index == 0:
                    write_mode = "w"
                    
                # determine if first row is header row and should be skipped in the write for any file other than the first
                if file_index != 0:
                    lines = lines[1:]
                
                write_result = write_report_data_base(
                    file_name=os.path.join(folder_path, output_file_name),
                    header=[],
                    data=lines,
                    write_type=write_mode,
                    delimiter=delimiter,
                    quoting=quoting
                )
                
                return_value.update(write_result)
            except Exception as e:
                return_value.update_sep(False, "File: {} failed to combine with exception: {}".format(file_, e))
        
    except Exception as e:
        return_value.update_sep(
            False, "Failed to combine files with exception: {}".format(e)
        )
    return return_value


def _format_headers(headers_in_file, file):
    """
    Replace any empty strings in header row with unique values based on the file name and a counter.

    :param headers_in_file: list of header entries
    :type headers_in_file: [str]
    :param file: Fully qualified file name
    :type file: str

    :return: Header row
    :rtype:[str]
    """

    file_name = get_file_name_without_ext(file)
    empty_header_counter = 0
    for i in range(len(headers_in_file)):
        # reformat any empty headers to be unique
        if headers_in_file[i] == "":
            headers_in_file[i] = file_name + ".Empty." + str(empty_header_counter)
            empty_header_counter = empty_header_counter + 1
    return headers_in_file


def combine_files_header_independent(
    folder_path,
    file_prefix="",
    file_suffix="",
    file_extension=".txt",
    output_file_name="result.txt",
    overwrite_existing=False,
    delimiter=",",
):
    """
    Used to combine report files into one file, files may have different number / named columns.

    Columns which are unique to some files will have as a value 'N/A' in files where those columns do not exist.

    :param folder_path: Folder path from which to get files to be combined and to which the combined file will be saved.
    :type folder_path: str
    :param file_prefix: Filter: File name starts with this value
    :type file_prefix: str
    :param file_suffix: Filter: File name ends with this value.
    :type file_suffix: str
    :param file_extension: Filter: File needs to have this file extension
    :type file_extension: str, format '.extension'
    :param out_put_file_name: The file name of the combined file, defaults to 'result.txt'
    :type out_put_file_name: str, optional
    :param overwrite_existing: Will overwrite an existing output file if set to True, defaults to False ( append to existing output file)
    :type overwrite_existing: bool, optional
    :param delimiter: The delimiter used in the files (e.g., ',' for CSV, '\t' for tab-separated), defaults to ','
    :type delimiter: str, optional

    :return:
        Result class instance.

        - result.status (bool) True if file content was combined without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will be complete list of rows appended.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()
    try:
        file_list = glob.glob(
            folder_path + "\\" + file_prefix + "*" + file_suffix + file_extension
        )

        # build list of unique headers
        headers = get_unique_headers_base(file_list, delimiter)
        combined_file_name = os.path.join(folder_path, output_file_name)

        # loop over files to be combined
        file_counter = 0
        for file in file_list:
            line_counter = 0
            column_mapper = []

            lines_result = read_column_based_text_file_base(
                file_path=file, delimiter=delimiter
            )
            if lines_result.status is False:
                raise Exception(
                    "Failed to read file: {} with {}".format(file, lines_result.message)
                )

            # get the lines read from the file
            lines = lines_result.result

            lines_to_be_transferred = []
            for line in lines:
                # read the headers in file
                if line_counter == 0:
                    # replace any empty strings in header
                    headers_in_file = _format_headers(line, file)
                    # match up unique headers with headers from this file
                    # build header mapping
                    for unique_header in headers:
                        if unique_header in headers_in_file:
                            column_mapper.append(headers_in_file.index(unique_header))
                        else:
                            column_mapper.append(-1)

                # ensure unique header is written to file
                if file_counter == 0 and line_counter == 0:
                    lines_to_be_transferred.append(headers)
                elif line_counter != 0:
                    padded_row = []
                    for cm in column_mapper:
                        if cm == -1:
                            # this column does not exist in this file
                            padded_row.append("N/A")
                        elif cm > len(line):
                            # less columns in file than mapper index (should'nt happen??)
                            padded_row.append("index out of bounds")
                        else:
                            padded_row.append(line[cm])
                    lines_to_be_transferred.append(padded_row)
                    # debug
                    return_value.result.append(padded_row)
                line_counter += 1

            # determine write type, default is append
            write_type = "a"
            # if overwrite existing is set, write type is write for the first file only!
            if file_counter == 0 and overwrite_existing:
                write_type = "w"
            # write file data to combined file
            result_write = write_report_data_base(
                combined_file_name,
                header=[],
                data=lines_to_be_transferred,
                write_type=write_type,
                delimiter=delimiter,
            )
            # keep track of what happened
            return_value.update(result_write)
            file_counter += 1
        return_value.append_message(
            "{} Files combined successfully.".format(file_counter)
        )
    except Exception as e:
        return_value.update_sep(
            False, "Failed to combine files with exception: {}".format(e)
        )
    return return_value


def append_to_file(
    source_file,
    append_file,
    ignore_first_row=False,
    delimiter=",",
    quoting=csv.QUOTE_MINIMAL,
):
    """
    Appends one text file to another.

    Assumes:

        - same number of headers (columns) in both files.
        - files are encoded in UTF-8!

    :param source_file: The fully qualified file path of the file to which the other file will be appended.
    :type source_file: str
    :param append_file: The fully qualified file path of the file to be appended.
    :type append_file: str
    :param ignore_first_row: If True, first row of append file will not be appended to source file.( Assumed its a header row )
    :type ignore_first_row: bool
    :param delimiter: The delimiter used in the files (e.g., ',' for CSV, '\t' for tab-separated), defaults to ','
    :type delimiter: str, optional
    :param quoting: The quoting option for the CSV writer, defaults to csv.QUOTE_MINIMAL
    :type quoting: int, optional

    :return:
        Result class instance.

        - result.status (bool) True if file content was appended without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will be an empty list.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()
        
    try:
        # read file to append into memory...hopefully will never get in GB range in terms of file size
        lines_result = read_column_based_text_file_base(
            append_file, delimiter=delimiter
        )
        if lines_result.status is False:
            return_value.update_sep(
                False,
                "Failed to read file: {} with {}".format(
                    append_file, lines_result.message
                ),
            )
            return return_value

        # get the lines read from the file depending on whether the first row is to be ignored
        lines = []
        if ignore_first_row:
            # remove the first row from the lines to be appended
            lines = lines_result.result[1:]
        else:
            # get the lines from the file
            lines = lines_result.result
        
        # write data to file
        write_result = write_report_data_base(
            file_name=source_file,
            header=[],
            data=lines,
            write_type="a",
            delimiter=delimiter,
            quoting=quoting,
        )

        return_value.update(write_result)
        
    except Exception as e:
        return_value.update_sep(
            False, "Failed to append file with exception: {}".format(e)
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Revit category graphics helper functions for project files.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

import duHast.Utilities.Objects.result as res
from duHast.Revit.LinePattern.line_patterns import get_all_line_patterns
from duHast.Revit.Categories.categories_model import get_categories_in_model
from duHast.Revit.Categories.Objects.object_style import ObjectStyle

from duHast.Revit.Common.Objects.Data.line_projection import LineProjection
from duHast.Revit.Common.Objects.Data.line_cut import LineCut

from duHast.Revit.LinePattern.Data.Objects.line_pattern_settings import (
    LinePatternSettings,
)
from duHast.Revit.Views.Utility.convert_revit_override_to_data import (
    _get_name_from_pattern_id,
)
from duHast.Revit.Common.Utility.revit_to_data_conversion import to_colour
from duHast.Revit.Common.transaction import in_transaction
from duHast.Revit.Materials.materials import get_material_by_id


from Autodesk.Revit.DB import (
    Category,
    Color,
    Document,
    ElementId,
    GraphicsStyleType,
    Transaction,
)


def set_cut_style(doc, category, line_cut_style):
    """
    Sets line cut settings for a category object.

    :param doc: A Revit Document object
    :type doc: Autodesk.Revit.DB.Document
    :param category: A Revit Category object
    :type category: Autodesk.Revit.DB.Category
    :param line_patterns: All line patterns in the model
    :type line_patterns: [AutoDesk.Revit.LinePatternElement]
    :param line_cut_style: An overrides projections object instance.
    :type line_cut_style: :class:`.LineCut`
    """

    return_value = res.Result()
    try:

        # check if doc is a Revit Document object
        if isinstance(doc, Document) == False:
            raise TypeError("doc must be a Revit Document object")

        # check if category is a Revit Category object
        if isinstance(category, Category) == False:
            raise TypeError("category must be a Revit Category object")

        # check if line_cut_style is a LineCut object
        if isinstance(line_cut_style, LineCut) == False:
            raise TypeError("line_cut_style must be a LineCut object")

        # start
        def action():
            action_return_value = res.Result()
            try:
                # set cut line settings
                category.SetLineWeight(line_cut_style.weight, GraphicsStyleType.Cut)
                new_colour = Color(
                    line_cut_style.colour.red,
                    line_cut_style.colour.green,
                    line_cut_style.colour.blue,
                )

                # set the colour value
                category.LineColor = new_colour

                # set line pattern by id
                # use projection as graphics style since cut is not available (?)
                category.SetLinePatternId(
                    ElementId(line_cut_style.line_pattern_settings.id),
                    GraphicsStyleType.Projection,
                )
                action_return_value.update_sep(
                    True,
                    "Successfully updated cut style of category: {}".format(
                        category.Name
                    ),
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to set cut style of category {} with exception: {}".format(
                        category.Name, e
                    ),
                )
            return action_return_value

        transaction = Transaction(doc, "Updating cut style: {}".format(category.Name))
        update_colour = in_transaction(transaction, action)
        return_value.update(update_colour)
    except Exception as e:
        return_value.update_sep(
            "failed to update cut style with exception: {}".format(e)
        )
    return return_value


def get_cut_style(category, line_patterns):
    """
    Retrieves line cut settings from a category object.

    :param category: A Revit Category object
    :type category: Autodesk.Revit.DB.Category
    :param line_patterns: All line patterns in the model
    :type line_patterns: [AutoDesk.Revit.LinePatternElement]

    :return: An overrides projections object instance.
    :rtype: :class:`.LineCut`
    """

    # start
    # cut line settings
    line_cut_style = LineCut()
    line_cut_style.colour = to_colour(category.LineColor)
    line_cut_style.weight = category.GetLineWeight(GraphicsStyleType.Cut)
    # basic line pattern details
    # using projection as graphics style since cut is not available
    line_pat_settings = LinePatternSettings(
        name=_get_name_from_pattern_id(
            id=category.GetLinePatternId(GraphicsStyleType.Projection),
            pattern_list=line_patterns,
        ),
        id=category.GetLinePatternId(GraphicsStyleType.Projection).IntegerValue,
    )
    # store line pattern details
    line_cut_style.line_pattern_settings = line_pat_settings

    return line_cut_style


def set_projection_style(doc, category, line_projection_style):
    """
    Sets line projection settings for a category object.

    :param doc: A Revit Document object
    :type doc: Autodesk.Revit.DB.Document
    :param category: A Revit Category object
    :type category: Autodesk.Revit.DB.Category
    :param line_patterns: All line patterns in the model
    :type line_patterns: [AutoDesk.Revit.LinePatternElement]
    :param line_projection_style: An overrides projections object instance.
    :type line_projection_style: :class:`.LineProjection`
    """

    return_value = res.Result()
    try:
        # check if doc is a Revit Document object
        if isinstance(doc, Document) == False:
            raise TypeError("doc must be a Revit Document object")

        # check if category is a Revit Category object
        if isinstance(category, Category) == False:
            raise TypeError("category must be a Revit Category object")

        # check if line_cut_style is a LineCut object
        if isinstance(line_projection_style, LineProjection) == False:
            raise TypeError("line_projection_style must be a LineProjection object")

        # start
        def action():
            action_return_value = res.Result()
            try:
                # set cut line settings
                category.SetLineWeight(
                    line_projection_style.weight,
                    GraphicsStyleType.Projection,
                )

                # set the color value
                new_colour = Color(
                    line_projection_style.colour.red,
                    line_projection_style.colour.green,
                    line_projection_style.colour.blue,
                )
                category.LineColor = new_colour

                # set line pattern by id
                category.SetLinePatternId(
                    ElementId(line_projection_style.line_pattern_settings.id),
                    GraphicsStyleType.Projection,
                )
                action_return_value.update_sep(
                    True,
                    "Successfully updated projection style of category: {}".format(
                        category.Name
                    ),
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to set projection style of category {} with exception: {}".format(
                        category.Name, e
                    ),
                )
            return action_return_value

        transaction = Transaction(
            doc, "Updating projection style: {}".format(category.Name)
        )
        update_colour = in_transaction(transaction, action)
        return_value.update(update_colour)
    except Exception as e:
        return_value.update_sep(
            "failed to update projection style with exception: {}".format(e)
        )
    return return_value


def get_projection_style(category, line_patterns):
    """
    Retrieves line projection settings from a category object.

    :param category: A Revit Category object
    :type category: Autodesk.Revit.DB.Category
    :param line_patterns: All line patterns in the model
    :type line_patterns: [AutoDesk.Revit.LinePatternElement]

    :return: An overrides projections object instance.
    :rtype: :class:`.LineProjection`
    """

    # start
    # projection line settings
    line_projection_style = LineProjection()
    line_projection_style.colour = to_colour(category.LineColor)
    line_projection_style.weight = category.GetLineWeight(GraphicsStyleType.Projection)
    # basic line pattern details
    line_pat_settings = LinePatternSettings(
        name=_get_name_from_pattern_id(
            id=category.GetLinePatternId(GraphicsStyleType.Projection),
            pattern_list=line_patterns,
        ),
        id=category.GetLinePatternId(GraphicsStyleType.Projection).IntegerValue,
    )
    # store line pattern details
    line_projection_style.line_pattern_settings = line_pat_settings

    return line_projection_style


def set_category_material_by_material_id(doc, category, material_id):
    """
    Sets the material of a category object.

    :param doc: A Revit Document object
    :type doc: Autodesk.Revit.DB.Document
    :param category: A Revit Category object
    :type category: Autodesk.Revit.DB.Category
    :param material_id: The id of the material to be set
    :type material_id: int
    """

    return_value = res.Result()
    try:
        # check if doc is a Revit Document object
        if isinstance(doc, Document) == False:
            raise TypeError("doc must be a Revit Document object")

        # check if category is a Revit Category object
        if isinstance(category, Category) == False:
            raise TypeError("category must be a Revit Category object")

        # check if material_id is an integer
        if isinstance(material_id, int) == False:
            raise TypeError("material_id must be an integer")

        # get material by id
        material = get_material_by_id(doc, ElementId(material_id))

        # check if material with this id exists in model
        if material == None and material_id != -1:
            raise ValueError(
                "Material with id {} not found in model".format(material_id)
            )

        # start
        def action():
            action_return_value = res.Result()
            try:
                category.Material = material
                action_return_value.update_sep(
                    True,
                    "Successfully updated material of category: {}".format(
                        category.Name
                    ),
                )
            except Exception as e:
                action_return_value.update_sep(
                    False,
                    "Failed to set material of category {} with exception: {}".format(
                        category.Name, e
                    ),
                )
            return action_return_value

        transaction = Transaction(doc, "Updating material: {}".format(category.Name))
        update_material = in_transaction(transaction, action)
        return_value.update(update_material)
    except Exception as e:
        return_value.update_sep(
            "failed to update material with exception: {}".format(e)
        )
    return return_value


def set_category_style(doc, category_style_source, category_style_target):
    """
    Sets category graphics settings for a Revit model.

    :param doc: A Revit Document object
    :type doc: Autodesk.Revit.DB.Document
    :param category_style_source: A list of category graphics settings
    :type category_style_source: [:class:`.ObjectStyle`]
    :param category_style_target: A list of category graphics settings
    :type category_style_target: [:class:`.ObjectStyle`]

    :return: A result object with the success status and the comparison result
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    try:
        # check if doc is a Revit Document object
        if isinstance(doc, Document) == False:
            raise TypeError("doc must be a Revit Document object")

        # check if category_style_source is a list of ObjectStyle instances
        if isinstance(category_style_source, ObjectStyle) is False:
            raise TypeError("category_style_source must be an ObjectStyle instance")

        # check if category_style_target is a list of ObjectStyle instances
        if isinstance(category_style_target, ObjectStyle) is False:
            raise TypeError("category_style_target must be an ObjectStyle instance")

        # get the revit category object
        target_category = Category.GetCategory(
            doc, ElementId(category_style_target.category_id)
        )

        # set cut style
        result_cut_style_update = set_cut_style(
            doc, target_category, category_style_source.line_cut
        )
        return_value.update(result_cut_style_update)

        # set projection style
        result_projection_style_update = set_projection_style(
            doc, target_category, category_style_source.line_projection
        )
        return_value.update(result_projection_style_update)

        # material
        result_material_update = set_category_material_by_material_id(
            doc, target_category, category_style_source.material_id
        )
        return_value.update(result_material_update)

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to set category graphics settings with exception: {}".format(e),
        )

    return return_value


def get_category_styles(doc):
    """
    Retrieves all category graphics settings from a Revit model.

    :param doc: A Revit Document object
    :type doc: Autodesk.Revit.DB.Document

    :return: A result object with the success status and the category graphics settings
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    # get all pattern element since they will be required in the category settings
    # get all line pattern in model
    line_pattern_elements = get_all_line_patterns(doc=doc)

    # retrieve all categories available in model
    categories_in_model = get_categories_in_model(doc=doc)

    try:
        # set up container for all object styles in model
        object_styles_in_model = []

        # get object styles for each category
        for model_cat in categories_in_model:

            # get the revit category object
            category = Category.GetCategory(doc, model_cat.id)

            # set up storage class
            object_style = ObjectStyle(
                main_category_name=model_cat.category_name,
                sub_category_name=model_cat.sub_category_name,
                category_id=model_cat.id.IntegerValue,
            )

            # get overrides for projection and cut
            object_style.line_projection = get_projection_style(
                category, line_pattern_elements
            )
            object_style.line_cut = get_cut_style(category, line_pattern_elements)

            # get the material id
            category_material = category.Material
            if category_material != None:
                object_style.material_id = category.Material.Id.IntegerValue
            else:
                object_style.material_id = -1

            # add to overall list
            object_styles_in_model.append(object_style)

        return_value.update_sep(
            True, "Successfully retrieved category graphics settings"
        )
        return_value.result = object_styles_in_model

    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to get category graphics settings with exception: {}".format(e),
        )

    return return_value


def get_all_used_line_patterns_ids_from_categories(categories):
    """
    Retrieves all line pattern ids as integer values from list of category graphics settings.

    :param categories: A list of category graphics settings
    :type categories: [:class:`.ObjectStyle`]

    :return: A list of line pattern ids
    :rtype: [int]
    """

    # check if categories is a list of ObjectStyle instances
    if (
        isinstance(categories, list) is False
        or all(isinstance(elem, ObjectStyle) for elem in categories) is False
    ):
        raise TypeError("categories must be a list of ObjectStyle instances")

    # start
    used_line_pattern_ids = []

    # get all line patterns used in categories
    for category in categories:
        used_patterns = category.get_all_used_line_patterns()
        for pattern_name, pattern in used_patterns.items():
            if pattern.id not in used_line_pattern_ids:
                used_line_pattern_ids.append(pattern.id)

    return used_line_pattern_ids


def get_not_equal_category_styles(category_styles_1, category_styles_2):
    """
    Compares two lists of category graphics settings.

    Notes:

    - The comparison is based on the main category name and sub category name.
    - if a style exists in list 1 but not in list 2 it will be reported as an equal style.
    - if a style exists in list 2 but not in list 1 it will not be reported.

    :param category_styles_1: A list of category graphics settings
    :type category_styles_1: [:class:`.ObjectStyle`]
    :param category_styles_2: A list of category graphics settings
    :type category_styles_2: [:class:`.ObjectStyle`]

    :return: A result object with the success status and the comparison result

            . result is a list of dictionaries with the main and sub category names as keys and the category graphics settings from both lists as values.
    :rtype: :class:`.Result`
    """

    return_value = res.Result()

    try:
        # check if categories is a list of ObjectStyle instances
        if (
            isinstance(category_styles_1, list) is False
            or all(isinstance(elem, ObjectStyle) for elem in category_styles_1) is False
        ):
            raise TypeError("category_styles_1 must be a list of ObjectStyle instances")
        if (
            isinstance(category_styles_2, list) is False
            or all(isinstance(elem, ObjectStyle) for elem in category_styles_2) is False
        ):
            raise TypeError("category_styles_2 must be a list of ObjectStyle instances")

        is_not_equal = []
        # outer loop
        for cat_style_1 in category_styles_1:
            matching_cat_style = [
                x
                for x in category_styles_2
                if x.main_category_name == cat_style_1.main_category_name
                and x.sub_category_name == cat_style_1.sub_category_name
            ]
            if len(matching_cat_style) == 0:
                return_value.append_message(
                    "Category style {} not found in second list".format(cat_style_1)
                )
            else:
                if cat_style_1 != matching_cat_style[0]:
                    is_not_equal.append(
                        {
                            "{}::{}".format(
                                cat_style_1.main_category_name,
                                cat_style_1.sub_category_name,
                            ): [cat_style_1, matching_cat_style[0]]
                        }
                    )

        return_value.update_sep(
            True, "Successfully compared category graphics settings"
        )
        return_value.result = is_not_equal
    except Exception as e:
        return_value.update_sep(
            False,
            "Failed to compare category graphics settings with exception: {}".format(e),
        )
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family Base data analysis module containing functions to find circular family referencing in extracted data.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

A circular reference is when a family A has a family B nested but family B has also family A nested.

Algorithm description:


- read families into data family objects
- process them ( to get the internal nesting set up)
- get the shortest path from each family
- check for duplicates in those paths ( may need to get unique path for name and category!)

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects.timer import Timer
from duHast.Utilities.Objects import result as res
from duHast.Revit.Family.Data.family_data_family_processor_utils import process_data

def find_circular_reference(family_data, result_list):
    """
    Function to find circular references in families.

    :param family_data: list of family data objects
    :type family_data: list
    :param result_list: list to store families with circular references
    :type result_list: list of tuples with two entries (0 index the family_data_family object, 1 index list of tuples with two entries in format, 0 index the nesting level as integer at which the circular nesting occurs, 1 index a string in format  family name :: family category )
    :return: list of tuples containing family with circular references at 0 and circular family data at 1
    """

    # loop over all family data
    for family in family_data:
        # process each family and check for circular references
        circular_families = family.has_circular_nesting()
        if len(circular_families) > 0:
            result_list.append((family, circular_families))
    return result_list


def check_families_have_circular_references(family_base_data_report_file_path):
    """
    Function to check if families have circular references.

    :param family_base_data_report_file_path: path to the family base data report file
    :type family_base_data_report_file_path: str
    :return: result object
    :rtype: Result
    """
    
    # read families into data family objects
    # check for duplicate family names in nesting tree

    return_value = res.Result()

    # set up a timer
    t_process = Timer()
    t_process.start()

    try:
        # load and process families
        families_processed_result = process_data(
        family_base_data_report_file_path=family_base_data_report_file_path,
            do_this=find_circular_reference,
        )

        # check if processing was successful, otherwise get out
        if families_processed_result.status == False:
            raise ValueError(families_processed_result.message)

        # get results
        families = []
        families_with_circular_nesting = []
        for nested_tuple in families_processed_result.result:
            families.append(nested_tuple[0])
            families_with_circular_nesting.extend(nested_tuple[1])

            # append messages debug
            return_value.append_message(families_processed_result.message)

            return_value.append_message(
                "{} Found: {} circular references in families.".format(
                    t_process.stop(), len(families_with_circular_nesting)
                )
            )

        # return families which have circular references
        if len(families_with_circular_nesting) > 0:
            return_value.result = families_with_circular_nesting

    except Exception as e:
        return_value.update_sep(
            False,
            "An error occurred while reading the families into data objects.{}".format(e),
            )

    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family report data utility module containing functions to read the data from file.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#

import os
from duHast.Revit.Family.Data.Objects.ifamily_data_storage import IFamilyDataStorage
from duHast.Revit.Family.Data.Objects.family_base_data_storage import (
    FamilyBaseDataStorage,
)
from duHast.Revit.Categories.Data.Objects.category_data_storage import (
    FamilyCategoryDataStorage,
)
from duHast.Revit.LinePattern.Data.Objects.line_pattern_data_storage import (
    FamilyLinePatternDataStorage,
)
from duHast.Revit.SharedParameters.Data.Objects.shared_parameter_data_storage import (
    FamilySharedParameterDataStorage,
)
from duHast.Revit.Warnings.Data.Objects.warnings_data_storage import (
    FamilyWarningsDataStorage,
)
from duHast.Revit.Family.Data.Objects.family_data_container import FamilyDataContainer
from duHast.Revit.Family.Data.Objects.family_data_family import FamilyDataFamily
from duHast.Revit.Family.Data.Objects.family_base_data_processor_defaults import (
    NESTING_SEPARATOR,
)

from duHast.Utilities.Objects.result import Result
from duHast.Utilities.files_io import file_exist
from duHast.Utilities.files_get import get_files_single_directory
from duHast.Utilities.files_csv import read_csv_file
from duHast.Utilities.directory_io import is_directory


DATA_CONVERSION = {
    FamilyBaseDataStorage.data_type: FamilyBaseDataStorage,
    FamilyCategoryDataStorage.data_type: FamilyCategoryDataStorage,
    FamilyLinePatternDataStorage.data_type: FamilyLinePatternDataStorage,
    FamilySharedParameterDataStorage.data_type: FamilySharedParameterDataStorage,
    FamilyWarningsDataStorage.data_type: FamilyWarningsDataStorage,
}


def convert_data_rows_to_data_storage(data_rows, target_type):
    """
    Convert the data rows into the target type of data storage object.

    :param data_rows: The data rows to be converted into the target type of data storage object.
    :type data_rows: list
    :param target_type: The target type of data storage object to convert the data rows into.
    :type target_type: IFamilyDataStorage
    :return: A Result object containing the list of IFamilyDataStorage objects if successful.
    :rtype: Result

    """

    return_value = Result()
    # flag checking whether an exception occurred when converting the data rows
    an_exception_occurred_counter = 0
    try:
        if not issubclass(target_type, IFamilyDataStorage):
            raise TypeError(
                "Invalid target type. Expected IFamilyDataStorage, got {}".format(
                    target_type.__name__
                )
            )
        # check if the data rows is a list
        if not isinstance(data_rows, list):
            raise TypeError(
                "Invalid data rows type. Expected list, got {}".format(type(data_rows))
            )

        # loop over all rows but the first (header row) extracted and set up objects
        for row in data_rows[1:]:
            try:
                # check if the row has the correct number of columns
                if len(row) != target_type.number_of_properties:
                    raise ValueError("Invalid data row{}".format(row))
                # initialise the storage object without the data type
                dummy = target_type(*row[1:])
                return_value.result.append(dummy)
            except Exception as e:
                return_value.append_message(
                    "Failed to convert data row with exception: {}".format(e)
                )
                an_exception_occurred_counter = an_exception_occurred_counter + 1

        # check if an exception occurred
        if an_exception_occurred_counter == 0:
            return_value.update_sep(
                True,
                "Successfully converted rows: {} into {}".format(
                    len(data_rows) - 1, target_type.__name__
                ),
            )
        else:
            return_value.update_sep(
                False,
                "Failed to convert rows into {}: {} rows failed to convert".format(
                    target_type.__name__, an_exception_occurred_counter
                ),
            )

    except Exception as e:
        return_value.update_sep(
            False, "Failed to convert rows into {}: {}".format(target_type.__name__, e)
        )
    return return_value


def read_data_from_file_into_storage(file_path, data_type):
    """
    Read the data from the file and return a list of IFamilyDataStorage objects.

    :param file_path: The path to the file to read the data from.
    :type file_path: str
    :param data_type: The type of data storage the file data is to be converted to.
    :type data_type: IFamilyDataStorage
    :return: A Result object containing the list of IFamilyDataStorage objects if successful.
    :rtype: Result

    """

    return_value = Result()
    try:
        # check what was past in
        if not isinstance(file_path, str):
            raise TypeError(
                "Invalid file path type. Expected str, got {}".format(type(file_path))
            )

        if not issubclass(data_type, IFamilyDataStorage):
            raise TypeError(
                "Invalid data type. Expected IFamilyDataStorage, got {}".format(
                    type(data_type)
                )
            )

        # Check if the file exists
        if not os.path.exists(file_path):
            raise FileNotFoundError("File not found: {}".format(file_path))

        # Read the data from the file
        data_result = read_csv_file(file_path)
        if data_result.status is False:
            raise Exception("Error reading data file: {}".format(data_result.message))
        
        data = data_result.result

        # Check if the data is empty
        if not data or len(data) <= 1:
            raise ValueError("Empty data in the file: {}".format(file_path))

        # check data type is what we expect
        # first column in second row contains the report data type
        if len(data[1]) == 0:
            raise ValueError("Data type missing in the file.")
        else:
            if not (data[1][0] == data_type.data_type):
                raise TypeError(
                    "Invalid data type.Got: {}, expected {}.".format(
                        data[1][0], data_type.data_type
                    )
                )

        # convert rows read into data storage objects
        data_storage_conversion_result = convert_data_rows_to_data_storage(
            data_rows=data, target_type=data_type
        )
        # check what came back before adding to converted data
        if data_storage_conversion_result.status:
            return_value.result = data_storage_conversion_result.result

        # just append the log message
        return_value.append_message(data_storage_conversion_result.message)

    except Exception as e:
        return_value.update_sep(
            False, "Failed to read data file with exception: {}".format(e)
        )

    return return_value


def read_family_base_data(file_path):
    """
    Read the family base data from the file and return a list of FamilyBaseDataStorage objects.

    :param file_path: The path to the file to read the data from.
    :type file_path: str
    :return: A Result object containing the list of FamilyBaseDataStorage objects if successful.
    :rtype: Result

    """

    return_value = Result()
    try:

        return_value = read_data_from_file_into_storage(
            file_path=file_path, data_type=FamilyBaseDataStorage
        )
    except Exception as e:
        return_value.update_sep(
            False, "Failed to read data file with exception: {}".format(e)
        )

    return return_value


def read_family_category_base_data(file_path):
    """
    Read the family category data from the file and return a list of FamilyCategoryDataStorage objects as part of a result object

    :param file_path: The path to the file to read the data from.
    :type file_path: str
    :return: A Result object containing the list of FamilyCategoryDataStorage objects if successful.
    :rtype: Result

    """
    return_value = Result()
    try:

        return_value = read_data_from_file_into_storage(
            file_path=file_path, data_type=FamilyCategoryDataStorage
        )
    except Exception as e:
        return_value.update_sep(
            False, "Failed to read data file with exception: {}".format(e)
        )

    return return_value


def read_family_line_pattern_base_data(file_path):
    """
    Read the family line pattern data from the file and return a list of FamilyLinePatternDataStorage objects as part of a result object

    :param file_path: The path to the file to read the data from.
    :type file_path: str
    :return: A Result object containing the list of FamilyLinePatternDataStorage objects if successful.
    :rtype: Result

    """
    return_value = Result()
    try:

        return_value = read_data_from_file_into_storage(
            file_path=file_path, data_type=FamilyLinePatternDataStorage
        )
    except Exception as e:
        return_value.update_sep(
            False, "Failed to read data file with exception: {}".format(e)
        )

    return return_value


def read_family_shared_parameter_data(file_path):
    """
    Read the family shared parameter data from the file and return a list of FamilySharedParameterDataStorage objects as part of a result object

    :param file_path: The path to the file to read the data from.
    :type file_path: str
    :return: A Result object containing the list of FamilySharedParameterDataStorage objects if successful.
    :rtype: Result

    """
    return_value = Result()
    try:

        return_value = read_data_from_file_into_storage(
            file_path=file_path, data_type=FamilySharedParameterDataStorage
        )
    except Exception as e:
        return_value.update_sep(
            False, "Failed to read data file with exception: {}".format(e)
        )

    return return_value


def read_family_warnings_data(file_path):
    """
    Read the family warnings data from the file and return a list of FamilyWarningsDataStorage objects as part of a result object

    :param file_path: The path to the file to read the data from.
    :type file_path: str
    :return: A Result object containing the list of FamilyWarningsDataStorage objects if successful.
    :rtype: Result

    """
    return_value = Result()
    try:

        return_value = read_data_from_file_into_storage(
            file_path=file_path, data_type=FamilyWarningsDataStorage
        )
    except Exception as e:
        return_value.update_sep(
            False, "Failed to read data file with exception: {}".format(e)
        )

    return return_value


def read_data_into_family_containers(path_to_data):
    """
    Get all csv files in directory provided and attempts to read them into varies lists of data storage objects:

    - FamilyBaseDataStorage
    - FamilyCategoryDataStorage
    - FamilyLinePatternDataStorage
    - FamilySharedParameterDataStorage
    - FamilyWarningsDataStorage

    These are then added to a family container object and returned. The family container object contains all storage instances of a specific root or nested family
    identified by a unique family name nesting path + family category nesting path

    Note: The content of multiple csv files of the same data type will be combined into a single list of objects.

    :param path_to_data: The path to the directory containing the csv files or fully qualified file path to a single data file.
    :type directory_path: str

    :return: A Result object containing the list of Family Containers objects if successful.
    :rtype: Result (list of FamilyDataContainer)
    """

    files = []
    return_value = Result()
    try:
        # check if path_to_data is a string
        if isinstance(path_to_data, str) == False:
            raise TypeError(
                "Invalid path type. Expected str, got {}".format(type(path_to_data))
            )

        # check if past in path is a directory or file
        # if it is a directory, get all csv files in the directory
        if is_directory(path_to_data):
            # get all csv files in the directory
            files = get_files_single_directory(
                folder_path=path_to_data,
                file_prefix="",
                file_suffix="",
                file_extension=".csv",
            )
        else:
            # if it is a file, check if it is a csv file
            if path_to_data.endswith(".csv"):
                # check if file exists
                if file_exist(path_to_data) == False:
                    raise FileNotFoundError("File not found: {}".format(path_to_data))
                else:
                    files.append(path_to_data)

        # read the data from each file into rows
        data_read = []
        for data_file in files:
            try:
                # attempt to read the data from the file
                data_result = read_csv_file(data_file)
                if data_result.status is False:
                    raise Exception(
                        "Error reading data file: {}".format(data_result.message)
                    )
                data = data_result.result
                
                if not data or len(data) <= 1:
                    raise ValueError("Empty data in the file: {}".format(data_file))
                else:
                    data_read.append(data)
            except Exception as e:
                return_value.append_message(
                    "Failed to read data file with exception: {}".format(e)
                )
        return_value.append_message("Read {} files".format(len(data_read)))

        # convert the data rows into storage objects depending on the data type
        # this will end up containing lists of storage objects, one list per file read
        data_converted = []
        for data in data_read:
            # check the first entry in the second row, since it contains the storage data type
            if len(data[1]) == 0:
                raise ValueError("Data type missing in the file.")
            else:
                if not (data[1][0] in DATA_CONVERSION.keys()):
                    raise TypeError("Invalid data type.Got: {}.".format(data[1][0]))
                else:
                    # convert row into storage object depending on the data type
                    data_type_name = data[1][0]
                    data_conversion_type = DATA_CONVERSION[data_type_name]
                    data_storage_conversion_result = convert_data_rows_to_data_storage(
                        data_rows=data, target_type=data_conversion_type
                    )
                    # check what came back before adding to converted data
                    if data_storage_conversion_result.status:
                        data_converted.append(data_storage_conversion_result.result)
                    # just append the log message
                    return_value.append_message(data_storage_conversion_result.message)

        return_value.append_message(
            "Converted {} data entries".format(len(data_converted))
        )
        # group storage containers by family root path and category root path identifying unique families
        containers_grouped_by_family_data = {}
        # loop over all storage lists and assign data to container
        for storage_list in data_converted:
            for storage_instance in storage_list:
                if (
                    storage_instance.root_name_path
                    + storage_instance.root_category_path
                    not in containers_grouped_by_family_data
                ):
                    # set up a new container for this family
                    container_instance = FamilyDataContainer()
                    # add the storage
                    container_instance.add_data_storage(storage_instance)
                    containers_grouped_by_family_data[
                        storage_instance.root_name_path
                        + storage_instance.root_category_path
                    ] = container_instance
                else:
                    # add the storage to the existing container
                    containers_grouped_by_family_data[
                        storage_instance.root_name_path
                        + storage_instance.root_category_path
                    ].add_data_storage(storage_instance)

        # add list of containers to return object
        for key, value in containers_grouped_by_family_data.items():
            return_value.result.append(value)

    except Exception as e:
        return_value.update_sep(
            False, "Failed to read data with exception: {}".format(e)
        )
    return return_value


def _get_root_families(families):
    """
    Returns all families which root families from past in list

    :param families: _description_
    :type families: _type_
    :return: _description_
    :rtype: _type_
    """
    root_families = []

    # do a type check
    if isinstance(families, list) == False:
        raise TypeError(
            "families needs to be of type list. Got: {} instead".format(type(families))
        )

    for family in families:
        # do a type check
        if isinstance(family, FamilyDataFamily) == False:
            raise TypeError(
                "All instances in families needs to be of type FamilyDataFamily. Got: {} instead".format(
                    type(family)
                )
            )

        if family.is_root_family:
            root_families.append(family)
    return root_families


def _assign_nested_families_to_root_families(root_families, families):
    """
    Adds nested families to their respective root families ( root family here is the top most family in the nesting tree)

    :param root_families: List of roo
    :type root_families: _type_
    :param families: _description_
    :type families: _type_
    :raises ValueError: _description_
    :return: _description_
    :rtype: _type_
    """
    for family in families:
        # check if this is a nested family ( not a root family )
        if family.is_root_family == False:
            # get the root family name and category
            root_name_path_chunks = family.family_nesting_path.split(NESTING_SEPARATOR)
            compare_family_name = root_name_path_chunks[0]
            category_name_path_chunks = family.family_category_nesting_path.split(
                NESTING_SEPARATOR
            )
            compare_family_category = category_name_path_chunks[0]

            # loop over root families and find the host
            found_root_family = False
            for root_family in root_families:
                if (
                    root_family.family_name == compare_family_name
                    and root_family.family_category == compare_family_category
                ):
                    root_family.add_nested_family_instance(family)
                    found_root_family = True
                    break

            # do a sanity check
            if found_root_family == False:
                raise ValueError(
                    "Cant find root family for {} {}".format(
                        compare_family_name, compare_family_category
                    )
                )
    return root_families


def read_data_into_families(path_to_data):
    """
    Read the data from the csv files in the directory and return a list of FamilyDataFamily objects.

    :param path_to_data: The path to the directory containing the csv files or fully qualified file path to single report csv file.
    :type path_to_data: str
    :return: A Result object containing the list of FamilyDataFamily objects if successful.
    :rtype: Result (list of FamilyDataFamily)

    """

    return_value = Result()
    families = []

    try:
        # first read reports into containers
        container_read_result = read_data_into_family_containers(path_to_data)
        return_value.update(container_read_result)

        # check if the read was successful
        if container_read_result.status == False:
            raise ValueError(
                "Failed to read data from: {} into family containers: {} ".format(
                    path_to_data, container_read_result.message
                )
            )

        # cycle over all containers and assign the data to the family objects
        # unique root families are identified by the following container properties
        # - the family name
        # - the family category

        containers = container_read_result.result
        # loop over containers and assign to families
        for container in containers:

            # check if the family is already in the list
            # compare_family_name = container.family_nesting_path
            # compare_family_category = container.family_category_nesting_path

            # if container.is_root_family:
            # if family is a root family use directly the containers family name and family category
            #    compare_family_name = container.family_nesting_path
            #    compare_family_category = container.family_category_nesting_path
            # else:
            # if a nested family get, need to get to the root family at the top of the nesting tree
            #    root_name_path_chunks = container.family_nesting_path.split(
            #        NESTING_SEPARATOR
            #    )
            #    compare_family_name = root_name_path_chunks[0]
            #    category_name_path_chunks = (
            #        container.family_category_nesting_path.split(NESTING_SEPARATOR)
            #    )
            #    compare_family_category = category_name_path_chunks[0]

            # check if the family is already in the list
            family_found = False
            for family in families:
                if (
                    family.family_name == container.family_name
                    and family.family_category == container.family_category
                    and family.family_nesting_path == container.family_nesting_path
                    and family.family_category_nesting_path
                    == container.family_category_nesting_path
                    and family.is_root_family == container.is_root_family
                ):
                    family_found = True
                    family.add_data_container(container)
                    return_value.append_message(
                        "Added container to family: {} - {} {} - {}".format(
                            container.family_name,
                            container.family_category,
                            container.family_nesting_path,
                            container.family_category_nesting_path,
                        )
                    )
                    break
            # if the family is not in the list, add it
            if family_found == False:
                new_family = FamilyDataFamily(
                    family_name=container.family_name,
                    family_category=container.family_category,
                    family_file_path=container.family_file_path,
                    family_nesting_path=container.family_nesting_path,
                    family_category_nesting_path=container.family_category_nesting_path,
                    is_root_family=container.is_root_family,
                )
                new_family.add_data_container(container)
                families.append(new_family)
                return_value.append_message(
                    "Added new family: {} - {} {} - {}".format(
                        container.family_name,
                        container.family_category,
                        container.family_nesting_path,
                        container.family_category_nesting_path,
                    )
                )

        # filter out root families
        root_families = _get_root_families(families=families)
        # assign nested families to root families
        families = _assign_nested_families_to_root_families(root_families, families)

    except Exception as e:
        return_value.update_sep(
            False, "Failed to convert containers into families: {}".format(e)
        )

    return_value.result = families
    return return_value



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
A base class for WPF view model error validation.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

Based on:

https://github.com/SingletonSean/wpf-tutorials/blob/master/ValidationMVVM/ViewModels/ErrorsViewModel.cs

"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from System.ComponentModel import INotifyDataErrorInfo,  PropertyChangedEventArgs



class ErrorsViewModel(INotifyDataErrorInfo):
    def __init__(self):
        """
        A base class for error validation in WPF view models.
        """
        # ini super class to allow multi inheritance in children!
        super(ErrorsViewModel, self).__init__()
        
        self._errors = {}
        self.propertyChangedHandlers = []
    

    def GetErrors(self, propertyName):
        """
        Returns the errors for the specified property.
        Implementation of INotifyDataErrorInfo.GetErrors.
        """
        
        if propertyName in self._errors:
            return self._errors[propertyName]
        return None

    def HasErrors(self):
        """
        Returns True if there are any errors in the view model.
        Implementation of INotifyDataErrorInfo.HasErrors.
        """
        return len(self._errors) > 0

    def add_ErrorsChanged(self, handler):
        """
        Adds a handler to the ErrorsChanged event.
        """
        self.propertyChangedHandlers.append(handler)

    def remove_ErrorsChanged(self, handler):
        """
        Removes a handler from the ErrorsChanged event.
        """
        self.propertyChangedHandlers.remove(handler)

    def RaiseErrorsChanged(self, propertyName):
        """
        Raises the ErrorsChanged event for the specified property.
        
        :param propertyName: The name of the property that has changed.
        :type propertyName: str
        """
        
        args = PropertyChangedEventArgs(propertyName)
        for handler in self.propertyChangedHandlers:
            handler(self, args)
        
    
    def ClearErrors(self, property_name):
        """
        Clears the errors for the specified property.
        
        :param property_name: The name of the property to clear the errors for.
        :type property_name: str
        """
        
        if property_name in self._errors:
            del self._errors[property_name]
            self.RaiseErrorsChanged(property_name)
    
    def ClearAllErrors(self):
        """
        Clears all errors.
        """
        
        self._errors.clear()
        self.RaiseErrorsChanged(None)
    
    def AddError(self, property_name, error_message):
        """
        Adds an error to the specified property.
        
        :param property_name: The name of the property to add the error to.
        :type property_name: str
        :param error_message: The error message.
        :type error_message: str
        """
        
        if property_name not in self._errors:
            self._errors[property_name] = []
        self._errors[property_name].append(error_message)
        self.RaiseErrorsChanged(property_name)


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a Revit level export to DATA class functions. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from Autodesk.Revit.DB import (
    Element,
)


from duHast.Revit.Levels.levels import get_levels_list_ascending
from duHast.Data.Objects.Collectors.data_level_building import DataLevelBuilding
from duHast.Revit.Exports.export_data import (
    get_model_data,
)
from duHast.Utilities.unit_conversion import convert_imperial_feet_to_metric_mm

def populate_data_level_object(doc, revit_level):
    """
    Returns a custom level data objects populated with some data from the revit model level past in.

    - level id
    - level name
    - level elevation

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param revit_level: A revit level instance.
    :type revit_level: Autodesk.Revit.DB.Level

    :return: A data level building object instance.
    :rtype: :class:`.DataLevelBuilding`
    """

    # set up data class object
    data_level_building = DataLevelBuilding()
    
    # id
    data_level_building.id = revit_level.Id.IntegerValue

    # name
    data_level_building.name = Element.Name.GetValue(revit_level)

    # elevation (converted to mm)
    data_level_building.elevation = convert_imperial_feet_to_metric_mm(revit_level.Elevation)

    # get the model name
    model = get_model_data(doc=doc)
    data_level_building.revit_model = model

    return data_level_building
   

def get_all_level_data(doc, filter_family_names=[]):
    """
    Gets a list of level data objects for each level element in the model.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param filter_family_names: filter list of level names to be ignored
    :type filter_family_names: [str]

    :return: A list of data level instances.
    :rtype: list of :class:`.DataLevelBuilding`
    """

    all_level_data = []
    revit_levels = get_levels_list_ascending(doc=doc)
    for revit_level in revit_levels:
        # check if level is in filter list and if so ignore it
        if len(filter_family_names) > 0:
            if Element.Name.GetValue(revit_level) in filter_family_names:
                continue
        level_data = populate_data_level_object(doc, revit_level)
        if level_data is not None:
            all_level_data.append(level_data)
    return all_level_data



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Helper functions attempting to use .net library to read a text file whilst avoiding encoding errors. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import os

from duHast.Utilities.Objects.result import Result
from duHast.Revit.NetSupport.dll_names import FILE_WRAPPER 

# load the wrapper dll from the libs folder
# the dll is located in the libs folder of the extension, which is one level up from the current file's directory
current_directory = os.path.dirname(__file__)
parent_directory = os.path.dirname(current_directory)
dll_path = os.path.join(parent_directory, "lib", FILE_WRAPPER)
clr.AddReference(dll_path)

# import the ReadFromFile class from the CSVHelperWrapper namespace
from duHastNet.FileIOWrapper import ReadFromColumnBasedTextFile, ReadFromTextFile

from duHast.Utilities.files_io import (
    get_file_name_without_ext,
)

def _read_text_file(file_path, delimiter=",", row_count=None):
    """
    Private helper function to read data from a text file using the CFileIOWrapper library.
    
    :param file_path: Path to the text file.
    :type file_path: str
    :param delimiter: Delimiter used in the text file. Default is ",".
    :type delimiter: str
    :param row_count: Number of rows to read. If None, reads the entire file.
    :type row_count: int or None
    
    :return:
        Result class instance.

        - result.status (bool) True if file was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain a list of nested lists of string representing the data read from the file.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
        - result.result will be an empty list.
    :rtype: :class:`.Result`
    """
    return_value = Result()
    
    try:
        # Validate input types
        if not isinstance(file_path, str):
            raise ValueError("file_path must be a string")
        if not isinstance(delimiter, str):
            raise ValueError("delimiter must be a string")
        if row_count is not None and not isinstance(row_count, int):
            raise ValueError("row_count must be an integer or None")
        
        # Instantiate the ReadFromFile class from the FileIO library
        reader_net = ReadFromColumnBasedTextFile()
        
        # Read the file based on row_count
        if row_count is None:
            result_list = reader_net.ReadFromTextFile(file_path, delimiter=delimiter)
        else:
            result_list = reader_net.ReadRowsFromTextFile(file_path, delimiter=delimiter, rowCount=row_count)
        
        # Handle empty result or errors
        if result_list.Count == 0:
            if reader_net.ErrorHistory.Count == 0:
                return_value.update_sep(True, "File was empty")
            else:
                for message in reader_net.ErrorHistory:
                    return_value.append_message(message)
                return_value.status = False
            return return_value
        
        # Convert .NET list to Python list
        for i in range(result_list.Count):
            nested_list = []
            for j in range(result_list[i].Count):
                nested_list.append(result_list[i][j])
            return_value.result.append(nested_list)
        
        return_value.update_sep(True, "Text file read from {} with status: {}".format(file_path, True))
        return return_value

    except Exception as e:
        return_value.update_sep(False, "Error reading text file: {}".format(e))
        return return_value


def read_from_delimited_text_file(file_path, delimiter=","):
    """
    Reads all data from a text file using the FileIOWrapper library.
    """
    return _read_text_file(file_path, delimiter)


def get_first_row_from_delimited_text_file(file_path, delimiter=","):
    """
    Reads the first row from a text file using the FileIOWrapper library.
    """
    return _read_text_file(file_path, delimiter, row_count=1)


def get_unique_headers(files, delimiter=","):
    """
    Gets a list of alphabetically sorted headers retrieved from text files.
    Assumes:

    - first row in each file is the header row

    :param files: List of file path from which the headers are to be returned.
    :type files: list of str
    :param delimiter: The delimiter used in the text file. Defaults to ','.
    :type delimiter: str, optional
    :return: List of headers.
    :rtype: list of str
    """

    headers_in_all_files = {}
    for f in files:
        
        # get unmodified row data and remove the next line character at the end
        data_result = get_first_row_from_delimited_text_file(f, delimiter=delimiter)
        if data_result.status is False:
            raise Exception("Failed to get headers from file: {}".format(f))

        # store the headers by file name
        headers_in_all_files[get_file_name_without_ext(f)] = data_result.result

    # create a list of unique headers
    headers_unique = []
    for header_by_file in headers_in_all_files:
        empty_header_counter = 0
        for header in headers_in_all_files[header_by_file]:
            # reformat any empty headers to be unique
            if header == "":
                header = header_by_file + ".Empty." + str(empty_header_counter)
                empty_header_counter = empty_header_counter + 1
            if header not in headers_unique:
                headers_unique.append(header)
    
    return sorted(headers_unique)


def read_non_delimited_text_file(file_path):
    """
    Read a non-column based text file.

    :param file_path: The fully qualified file path to the text file.
    :type file_path: str

    :return:
        Result class instance.

        - result.status (bool) True if file was read without an exception, otherwise False.
        - result.message contains log messages.
        - result.result will contain the data read from file.

        On exception:

        - result.status (bool) will be False.
        - result.message will contain exception message.
    :rtype: :class:`.Result`
    """

    return_value = Result()
    
    try:
        # Validate input types
        if not isinstance(file_path, str):
            raise ValueError("file_path must be a string")
        
        # Instantiate the ReadFromFile class from the FileIO library
        reader_net = ReadFromTextFile()
        
        # Read the file
        result_list = reader_net.ReadNonColumnBasedTextFile(file_path)
        
        # Handle empty result or errors
        if result_list.Count == 0:
            if reader_net.ErrorHistory.Count == 0:
                return_value.update_sep(True, "File was empty")
            else:
                for message in reader_net.ErrorHistory:
                    return_value.append_message(message)
                return_value.status = False
            return return_value
        
        # Convert .NET list to Python list
        for i in range(result_list.Count):
            return_value.result.append(result_list[i])
            
        return_value.update_sep(True, "Text file read from {} with status: {}".format(file_path, True))
        return return_value

    except Exception as e:
        return_value.update_sep(
            False, "Failed to read file: {} with exception: {}".format(file_path, e)
        )
    
    return return_value


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
IFC export settings class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import System
from duHast.Utilities.Objects import base


class IFCSettings(base.Base):

    supported_ifc_versions = {
        "Default": "The Autodesk Revit applications default export format. Note that this may change as the defaults change in the Revit user interface.",
        "IFCBCA": "IFC BCA file format. This is a certified variant of IFC 2x2 used for submitting files to the Singapore BCA ePlan Check Server.",
        "IFC2x2": "IFC 2x2 file format.",
        "IFC2x3": "IFC 2x3 file format.",
        "IFCCOBIE": "IFC GSA COBIE 2010 file format. This is a variant of IFC 2x3 used for submitting files that are COBIE 2010-complaint.",
        "IFC2x3CV2": "IFC 2x3 Coordination View 2.0 file format. This is a variant of IFC 2x3 used for exporting files using the Coordination View 2.0 model view.",
        "IFC4": "IFC 4 file format.",
        "IFC2x3FM": "IFC2x3 Extended FM Handover View",
        "IFC4RV": "IFC4 Reference View",
        "IFC4DTV": "IFC4 Design Transfer View",
        "IFC2x3BFM": "IFC2x3 Basic FM Handover View",
    }

    site_placement_options = {
        "SiteTransformBasis.Shared": "shared",
        "SiteTransformBasis.Site": "site",
        "SiteTransformBasis.Project": "project",
        "SiteTransformBasis.Internal": "internal",
        "SiteTransformBasis.ProjectInTN": "projectInTN",
        "SiteTransformBasis.InternalInTN": "internalInTN",
    }

    def __init__(
        self,
        name,
        ifc_version,
        space_boundaries,
        active_phase_id,
        active_view_id,
        export_base_quantities,
        split_walls_and_columns,
        visible_elements_of_current_view,
        use2_d_room_boundary_for_volume,
        use_family_and_type_name_for_reference,
        export_internal_revit_property_sets,
        export_ifc_common_property_sets,
        export_2d_elements,
        export_parts_as_building_elements,
        export_bounding_box,
        export_solid_model_rep,
        export_schedules_as_psets,
        export_user_defined_psets,
        export_user_defined_psets_file_name,
        export_linked_files,
        include_site_elevation,
        use_active_view_geometry,
        export_specific_schedules,
        tessellation_level_of_detail,
        store_ifc_guid,
        export_rooms_in_view,
        use_only_triangulation,
        include_steel_elements,
        cobie_company_info,
        cobie_project_info,
        use_type_name_only_for_ifc_type,
        use_visible_revit_name_as_entity_name,
        site_placement,
        selected_site,
        geo_ref_crs_name,
        geo_ref_crs_desc,
        geo_ref_epsg_code,
        geo_ref_geodetic_datum,
        geo_ref_map_unit,
        exclude_filter,
    ):

        """
        _summary_

        :param name: The name of the configuration.
        :type name:
        :param ifcVersion: The IFCFileFormat of the configuration.
        :type ifcVersion:
        :param spaceBoundaries: The level of space boundaries of the configuration.
        :type spaceBoundaries:
        :param activePhaseId: The phase of the document to export.
        :type activePhaseId:
        :param exportBaseQuantities: Whether or not to include base quantities for model elements in the export data. Base quantities are generated from model geometry to reflect actual physical quantity values, independent of measurement rules or methods.
        :type exportBaseQuantities:
        :param splitWallsAndColumns: Whether or not to split walls and columns by building stories.
        :type splitWallsAndColumns:
        :param visibleElementsOfCurrentView: True to export only the visible elements of the current view (based on filtering and/or element and category hiding). False to export the entire model.
        :type visibleElementsOfCurrentView:
        :param use2DRoomBoundaryForVolume: True to use a simplified approach to calculation of room volumes (based on extrusion of 2D room boundaries) which is also the default when exporting to IFC 2x2. False to use the Revit calculated room geometry to represent the room volumes (which is the default when exporting to IFC 2x3).
        :type use2DRoomBoundaryForVolume:
        :param useFamilyAndTypeNameForReference: True to use the family and type name for references. False to use the type name only.
        :type useFamilyAndTypeNameForReference:
        :param exportInternalRevitPropertySets: True to include the Revit-specific property sets based on parameter groups. False to exclude them.
        :type exportInternalRevitPropertySets:
        :param exportIFCCommonPropertySets: True to include the IFC common property sets. False to exclude them.
        :type exportIFCCommonPropertySets:
        :param export2DElements: True to include 2D elements supported by IFC export (notes and filled regions). False to exclude them.
        :type export2DElements:
        :param exportPartsAsBuildingElements: True to export the parts as independent building elements. False to export the parts with host element.
        :type exportPartsAsBuildingElements:
        :param exportBoundingBox: True to export bounding box. False to exclude them.
        :type exportBoundingBox:
        :param exportSolidModelRep: True to allow exports of solid models when possible. False to exclude them.
        :type exportSolidModelRep:
        :param exportSchedulesAsPsets: True to allow exports of schedules as custom property sets. False to exclude them.
        :type exportSchedulesAsPsets:
        :param exportUserDefinedPsets: True to allow user defined property sets to be exported. False to ignore them
        :type exportUserDefinedPsets:
        :param exportUserDefinedPsetsFileName: The name of the file containing the user defined Parameter Mapping Table to be exported.
        :type exportUserDefinedPsetsFileName:
        :param exportLinkedFiles: True to include links in export, otherwise False.
        :type exportLinkedFiles:
        :param includeSiteElevation: True to include IFCSITE elevation in the site local placement origin.
        :type includeSiteElevation:
        :param useActiveViewGeometry: True to use the active view when generating geometry. False to use default export options.
        :type useActiveViewGeometry:
        :param exportSpecificSchedules: True to export specific schedules
        :type exportSpecificSchedules:
        :param tessellationLevelOfDetail: Value indicating the level of detail to be used by tessellation. Valid values is between 0 to 1
        :type tessellationLevelOfDetail:
        :param storeIFCGUID: True to store the IFC GUID in the file after the export.  This will require manually saving the file to keep the parameter.
        :type storeIFCGUID:
        :param exportRoomsInView: True to export rooms if their bounding box intersect with the section box. If the section box isn't visible, then all the rooms are exported if this option is set.
        :type exportRoomsInView:
        :param useOnlyTriangulation: Value indicating whether tessellated geometry should be kept only as triangulation. (Note: in IFC4_ADD2 IfcPolygonalFaceSet is introduced that can simplify the coplanar triangle faces into a polygonal face. This option skip this)
        :type useOnlyTriangulation:
        :param includeSteelElements: Value indicating whether steel elements should be exported.
        :type includeSteelElements:
        :param cOBieCompanyInfo: COBie specific company information (from a dedicated tab)
        :type cOBieCompanyInfo:
        :param cOBieProjectInfo: COBie specific project information (from a dedicated tab)
        :type cOBieProjectInfo:
        :param activeViewId: Id of the active view.
        :type activeViewId: int
        :param useTypeNameOnlyForIfcType: Value indicating whether only the Type name will be used to name the IfcTypeObject
        :type useTypeNameOnlyForIfcType: bool
        :param useVisibleRevitNameAsEntityName: Value indicating whether the IFC Entity Name will use visible Revit Name
        :type useVisibleRevitNameAsEntityName: bool
        :param selectedSite: Selected Site name
        :type selectedSite: str
        :param sitePlacement: The origin of the exported file: either shared coordinates (Site Survey Point), Project Base Point, or internal coordinates.
        :type sitePlacement: str: SiteTransformBasis.Shared, SiteTransformBasis.Site, SiteTransformBasis.Project, SiteTransformBasis.Internal, SiteTransformBasis.ProjectInTN, SiteTransformBasis.InternalInTN
        :param geoRefCRSName: Projected Coordinate System Name
        :type geoRefCRSName: str
        :param geoRefCRSDesc: Projected Coordinate System Description
        :type geoRefCRSDesc: str
        :param geoRefEPSGCode: EPSG Code for the Projected CRS
        :type geoRefEPSGCode: str
        :param geoRefGeodeticDatum: The geodetic datum of the ProjectedCRS
        :type geoRefGeodeticDatum: str
        :param geoRefMapUnit: The Map Unit of the ProjectedCRS
        :type geoRefMapUnit: str
        :param excludeFilter: Exclude filter string (element list in an array, separated with semicolon ';')
        :type excludeFilter: str

        :raises Exception: _description_
        """

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(IFCSettings, self).__init__()

        self.name = name

        # check the IFC version
        if ifc_version in self.supported_ifc_versions:
            self.ifc_version = ifc_version
        else:
            raise Exception("Unsupported IFC version:" + ifc_version)

        self.space_boundaries = space_boundaries
        self.active_phase_id = active_phase_id
        self.active_view_id = active_view_id
        self.export_base_quantities = export_base_quantities
        self.split_walls_and_columns = split_walls_and_columns
        self.visible_elements_of_current_view = visible_elements_of_current_view
        self.use2_d_room_boundary_for_volume = use2_d_room_boundary_for_volume
        self.use_family_and_type_name_for_reference = (
            use_family_and_type_name_for_reference
        )
        self.export_internal_revit_property_sets = export_internal_revit_property_sets
        self.export_ifc_common_property_sets = export_ifc_common_property_sets
        self.export_2d_elements = export_2d_elements
        self.export_parts_as_building_elements = export_parts_as_building_elements
        self.export_bounding_box = export_bounding_box
        self.export_solid_model_rep = export_solid_model_rep
        self.export_schedules_as_psets = export_schedules_as_psets
        self.export_user_defined_psets = export_user_defined_psets
        self.export_user_defined_psets_file_name = export_user_defined_psets_file_name
        self.export_linked_files = export_linked_files
        self.include_site_elevation = include_site_elevation
        self.use_active_view_geometry = use_active_view_geometry
        self.export_specific_schedules = export_specific_schedules
        self.tessellation_level_of_detail = tessellation_level_of_detail
        self.store_ifc_guid = store_ifc_guid
        self.export_rooms_in_view = export_rooms_in_view
        self.use_only_triangulation = use_only_triangulation
        self.include_steel_elements = include_steel_elements
        self.cobie_company_info = cobie_company_info
        self.cobie_project_info = cobie_project_info
        self.use_type_name_only_for_ifc_type = (use_type_name_only_for_ifc_type,)
        self.use_visible_revit_name_as_entity_name = (
            use_visible_revit_name_as_entity_name
        )

        # check site placement option
        if site_placement in self.site_placement_options:
            self.site_placement = self.site_placement_options[site_placement]
        else:
            raise Exception("Unsupported site placement option:" + site_placement)

        self.selected_site = selected_site
        self.geo_ref_crs_name = geo_ref_crs_name
        self.geo_ref_crs_desc = geo_ref_crs_desc
        self.geo_ref_epsg_code = geo_ref_epsg_code
        self.geo_ref_geodetic_datum = geo_ref_geodetic_datum
        self.geo_ref_map_unit = geo_ref_map_unit
        self.exclude_filter = exclude_filter



# -*- coding: utf-8 -*-
"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains string functions for string re-formatting. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2025, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import re


def remove_currency_sign(number_string):
    """
    Remove a set of currency signs from a string.

    supported currency signs: $, €, £, ¥, ₹

    :param number_string: The string to remove the currency sign from.
    :type number_string: str

    :return: The string without the currency sign.
    :rtype: str
    """

    # Regular expression to match common currency signs at the start of the string
    pattern = r"^[\$\€\£\¥\₹]"

    # Remove the currency sign if present
    cleaned_string = re.sub(pattern, "", number_string)

    return cleaned_string


def replace_new_lines(input_string):
    """
    Replace all new line characters with a space and remove trailing spaces.

    :param input_string: The string to modify.
    :type input_string: str

    :return: The modified string.
    :rtype: str
    """

    # Replace all new line characters with a space
    modified_string = input_string.replace("\n", " ")
    # Remove trailing spaces
    return modified_string.rstrip()


def remove_trailing_characters_from_number_string(number_string):
    """
    Remove trailing characters from a number string.
    This function is used to remove unit strings from a number string.

    :param number_string: The string to remove trailing characters from.
    :type number_string: str

    :return: The number string without trailing characters.
    :rtype: str
    """

    # Check if p_value contains a number followed by a unit string (including special characters)
    number_unit_pattern = re.compile(r"^(\d+(\.\d+)?)\s*([^\d\s]+)$")
    match = number_unit_pattern.match(number_string)
    if match:
        # found a unit string, just return the number
        value = match.group(1)
        return value
    else:
        # No unit string found, just use the value as is
        return number_string



"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to Revit views. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

import clr
import System

# import common library modules
from duHast.Revit.Common import common as com
from duHast.Revit.Views.Utility.view_types import _get_view_types
from duHast.Revit.Views.schedules_revision_filter import (
    filter_revision_schedules,
    filter_schedules,
)
from duHast.Revit.Views.sheets import get_all_sheets
from duHast.Revit.Areas.areas import get_area_scheme_by_name, get_area_schemes, get_views_by_area_scheme_name
from duHast.Revit.Common.parameter_get_utils import (
    get_built_in_parameter_value,
    get_parameter_value_as_element_id,
)

# import Autodesk
from Autodesk.Revit.DB import (
    BuiltInParameter,
    ElementClassFilter,
    ElementId,
    FilteredElementCollector,
    View,
    ViewType,
)


def get_view_types(doc):
    """
    Returns all view family types in a model
    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :return: A filtered element collector.
    :rtype: Autodesk.Revit.DB.FilteredElementCollector
    """

    collector = _get_view_types(doc)
    return collector


def get_view_type_ids(doc):
    """
    Returns all view family type ids in a model

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: ids of view family types
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    ids = []
    col = get_view_types(doc)
    ids = com.get_ids_from_element_collector(col)
    return ids


def get_views_of_type(doc, view_type):
    """
    Gets all views in a model of a given type. Excludes templates.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view_type: Filter: the view type
    :type view_type: Autodesk.Revit.DB.ViewType

    :return: list of views
    :rtype: list of Autodesk.Revit.DB.View
    """

    views = []
    col = FilteredElementCollector(doc).OfClass(View)
    for v in col:
        if v.ViewType == view_type and v.IsTemplate == False:
            views.append(v)
    return views


# ----------------------------------------------------------------------------------------


def get_viewport_on_sheets(doc, sheets):
    """
    Get all view ports on sheets provided.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param sheets: list of sheets of which to return the view ports from.
    :type sheets: list of Autodesk.Revit.DB.ViewSheet

    :return: list of view ports
    :rtype: list of Autodesk.Revit.DB.Viewport
    """

    view_ports = []
    for sheet in sheets:
        try:
            viewport_ids = sheet.GetAllViewports()
            if viewport_ids != None:
                for viewport_id in viewport_ids:
                    viewport = doc.GetElement(viewport_id)
                    view_ports.append(viewport)
        except Exception as e:
            print("Get view ports on sheet: {} threw exception: {}".format(sheet, e))
    return view_ports


def get_views_in_model(doc, filter):
    """
    Gets all views in a model which are matching a filter and are:

    - not template views
    - not system browser
    - not project browser
    - not undefined
    - not Internal
    - not sheets


    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param filter: function checking view
    :type filter: func(view) returning a bool

    :return: list of views
    :rtype: list of Autodesk.Revit.DB.View
    """

    views = []
    col = FilteredElementCollector(doc).OfClass(View)
    for v in col:
        # filter out browser organization and other views which cant be deleted
        if (
            v.IsTemplate == False
            and filter(v) == True
            and v.ViewType != ViewType.SystemBrowser
            and v.ViewType != ViewType.ProjectBrowser
            and v.ViewType != ViewType.Undefined
            and v.ViewType != ViewType.Internal
            and v.ViewType != ViewType.DrawingSheet
        ):
            views.append(v)
    return views


def get_views_in_model_not_template(doc):
    """
    Gets all views in a model which are not templates and not driven by a template.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: list of views
    :rtype: list of Autodesk.Revit.DB.View
    """

    def filter_not_template(view):
        return view.IsTemplate == False

    return get_views_in_model(doc=doc, filter=filter_not_template)



def get_views_not_on_sheet(doc):
    """
    Gets all views not placed on a sheet. (Excludes schedules)

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document

    :return: A list of views which are currently not placed on a sheet.
    :rtype: list of Autodesk.Revit.DB.View
    """

    views_not_on_sheet = []
    # get all sheets
    sheets_in_model = get_all_sheets(doc)
    # get all viewPorts on sheets
    view_ports_on_sheets = get_viewport_on_sheets(doc, sheets_in_model)
    # get all views in model
    views_in_model = get_views_in_model(doc=doc, filter=filter_schedules)
    # check whether view has a viewport if not ... its not placed on a sheet
    for view_in_model in views_in_model:
        match = False
        for view_ports_on_sheet in view_ports_on_sheets:
            if view_ports_on_sheet.ViewId == view_in_model.Id:
                match = True
                break
        if match == False:
            views_not_on_sheet.append(view_in_model)
    return views_not_on_sheet


def get_view_phase_id(view):
    """
    Get the views phase id.

    Note if view does not support phase id and Invlaid ElementId (-1) is returned

    :param view: The view of which to return the phase id
    :type view: Autodesk.Revit.DB.View

    :return: An element id representing the phase id. If view does not support phases then an Invalid id (-1) will be returned.
    :rtype: list of Autodesk.Revit.DB.ElementId
    """

    # set up the default value ( no phase )
    return_value = ElementId.InvalidElementId

    # attempt to get the phase id from the view
    phase_id = get_built_in_parameter_value(
        element=view,
        built_in_parameter_def=BuiltInParameter.VIEW_PHASE,
        parameter_value_getter=get_parameter_value_as_element_id,
    )

    # check what came back and return accordingly
    if phase_id is not None:
        return phase_id
    else:
        return return_value


def get_area_scheme_of_view(doc, view):
    """
    Returns the area scheme an area plan is associated with.

    :param doc: Current Revit model document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view of which to return the area scheme
    :type view: Autodesk.Revit.DB.View
    :return: The area scheme if the view is of type AreaPlan otherwise None
    :rtype: str or None
    """
    if view.ViewType != ViewType.AreaPlan:
        return None
    
    area_schemes_in_model = get_area_schemes(doc=doc)

    for area_scheme in area_schemes_in_model:
        views_by_area_scheme = get_views_by_area_scheme_name(doc=doc, area_scheme_name=area_scheme.Name)
        for view_in_scheme in views_by_area_scheme:
            if view_in_scheme.Id == view.Id:
                return area_scheme
    
    return None


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of helper functions relating to py Revit element selection from list. 
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#



from Autodesk.Revit.DB import ElementId


def default_name_builder (element):
    """
    Deafult element name builder is used to build the element name shown in the UI by combining the element name and element  Id

    :param element: An element
    :type element: Autodesk.Revit.DB.Element
    :return: A name of the element.
    :rtype: str
    """
    key = "{} ({})".format(element.Name, element.Id)
    return key


def get_element_selection_from_user(doc, forms, element_getter, element_selection_description, multiselect = True, ui_element_name_builder = default_name_builder):
    """
    lists Elements provided by element getter function in UI and returns the users selection

    :param doc: Ther current Revit model.
    :type doc: Autodesk.Revit.DB.Document
    :param forms: pyRevit forms
    :type forms: _type_
    :param element_getter: Function accpeting the document as the only argument returning a list of elements or empty list.
    :type element_getter: _type_
    :param element_selection_description: Text to be displayes on button to prompt user
    :type element_selection_description: str
    
    :return: None if nothing was selected. Otherwise a list of element ids
    :rtype: None or [Autodesk.Revit.ElementId]
    """
    # set up return values
    element_names = []
    elements_by_name = {}

    # get all elements from the getter
    elements = element_getter(doc=doc)
    
    # check if we got any?
    if elements is None or len(elements) == 0:
        return None

    for element in elements:
        # check if this is an element id rather than an element
        if isinstance(element, ElementId):
            element = doc.GetElement(element)
        key = ui_element_name_builder(element=element)
        element_names.append(key)
        if(key in elements_by_name):
            print("Warning element {} exists twice in the model!".format(key))
        elements_by_name[key] = element
    
    elements_selected = None

    # check if we got any?
    if(len(element_names)==0):
        print("No elements left to display")
        return elements_selected

    # get the user to select the source ( returns a string)
    selection = forms.SelectFromList.show(sorted(element_names), button_name='{}'.format(element_selection_description), multiselect= multiselect)
    
    if(selection == None):
        return elements_selected
    else:
        elements_selected = []
        # check if string ( single selection) or list ( multiple selection)
        if(isinstance(selection, str)):
            elements_selected.append(elements_by_name[selection].Id)
            return elements_selected
        else:
            for element_key_name in selection:
                elements_selected.append(elements_by_name[element_key_name].Id)
            return elements_selected


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
Family data storage class.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This class is used to store data from families placed in a project environment.


"""

#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2024, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#

from duHast.Utilities.Objects import base

UNKNOWN_HOST_STATUS = "Host status could not be determined."

class FamilyReportData(base.Base):
    def __init__(self):

        # forwards all unused arguments
        # ini super class to allow multi inheritance in children!
        super(FamilyReportData, self).__init__()

        self.project_name = None
        self.family_category = None
        self.family_name = None
        self.family_type_name = None
        self.family_instances_placed = -1
        self._nested_families = [] #(set underlying list)
        self.is_host = False
        self._type_properties = [] #(set underlying list)
        self.is_shared = False


    @property
    def project_name(self):
        return self._project_name

    @project_name.setter
    def project_name(self, value):
        self._project_name = value

    @property
    def family_category(self):
        return self._family_category

    @family_category.setter
    def family_category(self, value):
        self._family_category = value

    @property
    def family_name(self):
        return self._family_name

    @family_name.setter
    def family_name(self, value):
        self._family_name = value

    @property
    def family_type_name(self):
        return self._family_type_name

    @family_type_name.setter
    def family_type_name(self, value):
        self._family_type_name = value

    @property
    def family_instances_placed(self):
        return self._family_instances_placed

    @family_instances_placed.setter
    def family_instances_placed(self, value):
        self._family_instances_placed = value

    @property
    def nested_families(self):
        return self._nested_families

    def add_nested_family(self, family_name):
        self._nested_families.append(family_name)
        # check if a family name was added or just a note
        if(family_name!=UNKNOWN_HOST_STATUS):
            # set the host attribute
            self.is_host = True

    @property
    def is_host(self):
        return self._is_host

    @is_host.setter
    def is_host(self, value):
        self._is_host = value

    @property
    def type_properties(self):
        return self._type_properties

    def add_type_property(self, property):
        self._type_properties.append(property)

       
    @property
    def is_shared(self):
        return self._is_shared

    @is_shared.setter
    def is_shared(self, value):
        self._is_shared = value
    

    def get_properties_as_list_str(self):
        """
        returns all property values  as a list of strings

        :return: A list of all property values.
        :rtype: [str]
        """

        data = []
        data.append(self.project_name)
        data.append(self.family_name)
        data.append(self.family_type_name)
        data.append(self.family_category)
        data.append(str(self.is_shared))
        data.append(str(self.is_host))

        # build nested family names
        nested_family_names = []
        for fam_name in self.nested_families:
            nested_family_names.append(fam_name)
        data.append(",".join(nested_family_names))

       
        # build type properties
        for type_p in self.type_properties:
            for property_name, property_value in type_p.items():
                if(isinstance(property_value, str)==False):
                    data.append(str(property_value))
                else:
                    data.append(property_value)
        
        data.append(str(self.family_instances_placed))
        return data

    def get_property_headers(self):
        """
        Builds a list of all proprty names.

        :return: A list of all property names.
        :rtype: [str]
        """

        data = []
        data.append("Project Name")
        data.append("Family Name")
        data.append("Family Type Name")
        data.append("Family Category")
        data.append("Is Family Shared")
        data.append("Is Family hosting shared families")
        data.append("Nested Family Names")
       
        # build type property names
        for type_p in self.type_properties:
            for property_name, property_value in type_p.items():
                data.append("Type Property: {}".format(property_name))
        
        data.append("Number Of Instances Placed")
        return data


"""
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
This module contains a number of functions around Revit 2D curve creation.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
"""
#
# License:
#
#
# Revit Batch Processor Sample Code
#
# BSD License
# Copyright 2023, Jan Christel
# All rights reserved.

# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

# - Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
# - Neither the name of the copyright holder nor the names of its contributors may be used to endorse or promote products derived from this software without specific prior written permission.
#
# This software is provided by the copyright holder "as is" and any express or implied warranties, including, but not limited to, the implied warranties of merchantability and fitness for a particular purpose are disclaimed.
# In no event shall the copyright holder be liable for any direct, indirect, incidental, special, exemplary, or consequential damages (including, but not limited to, procurement of substitute goods or services; loss of use, data, or profits;
# or business interruption) however caused and on any theory of liability, whether in contract, strict liability, or tort (including negligence or otherwise) arising in any way out of the use of this software, even if advised of the possibility of such damage.
#
#
#


from duHast.Utilities.Objects.result import Result
from duHast.Revit.Common import transaction as rTran

from Autodesk.Revit.DB import (
    Line,
    Transaction,
    XYZ
)

def draw_line_between_points(doc, view, start_point, end_point, transaction_manager=rTran.in_transaction):
    """
    Draws a line between two points in a Revit view.
    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param view: The view in which to draw the line.
    :type view: Autodesk.Revit.DB.View
    :param start_point: The start point of the line.
    :type start_point: Autodesk.Revit.DB.XYZ
    :param end_point: The end point of the line.
    :type end_point: Autodesk.Revit.DB.XYZ
    :return: Result object containing the status of the operation.
    :rtype: Result
    """

    # set up an action to be run in a transaction
    # to draw a 2D line
    # in the provided view
    def action():
        action_return_value = Result()
        try:
            line_to_draw = Line.CreateBound(start_point, end_point)
            dc = doc.Create.NewDetailCurve(view, line_to_draw)
            action_return_value.update_sep(True, "Line drawn")
            return action_return_value
        except Exception as e:
            action_return_value.update_sep(False, "Failed to draw line: {}".format(e))
            return action_return_value
       

    # if no transaction manager is provided, assume this is running in a transaction
    # started outside of this function
    if transaction_manager is None:
        return_value = action()
        return return_value
    
    # otherwise, use the provided transaction manager
    transaction = Transaction(doc, "Drawing line")
    return_value = transaction_manager(transaction, action)
    return return_value


def draw_2D_lines_on_bounding_box(doc, bounding_box, view,  transaction_manager=rTran.in_transaction):
    """
    Draw 2D lines on the bounding box of an element in view provided
    
    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param bounding_box: The bounding box of the element.
    :type bounding_box: Autodesk.Revit.DB.BoundingBoxXYZ
    :param view: The view in which to draw the lines.
    :type view: Autodesk.Revit.DB.View
    :param transaction_manager: The transaction manager to use.
    :type transaction_manager: function
    
    :return: Result object containing the status of the operation.
    :rtype: Result
    """

    return_value = Result()

    # get the corners of the bounding box
    min_x = bounding_box.Min.X
    min_y = bounding_box.Min.Y
    max_x = bounding_box.Max.X
    max_y = bounding_box.Max.Y

    # create a list of points for the lines
    points = [
        XYZ(min_x, min_y, 0),
        XYZ(max_x, min_y, 0),
        XYZ(max_x, max_y, 0),
        XYZ(min_x, max_y, 0),
        XYZ(min_x, min_y, 0)
    ]

    def action():
        action_return_value = Result()
        try:
            # draw the lines on the view
            for i in range(len(points) - 1):
                create_line_result = draw_line_between_points(doc, view, start_point=points[i], end_point=points[i + 1],  transaction_manager=None)
                action_return_value.update(create_line_result)
        except Exception as e:
            action_return_value.update_sep(False, "Failed to draw lines: {}".format(e))

        return action_return_value

    # use the provided transaction manager to draw the bounding box
    transaction = Transaction(doc, "Drawing bounding box")
    return_value = transaction_manager(transaction, action)

    return return_value


def draw_2D_lines_on_bounding_box_and_separate_point(doc, bounding_box, additional_point, view,  transaction_manager=rTran.in_transaction):
    """
    Draw 2D lines on the bounding box of an element in view provided and in addition lines from each corner of the bounding box to the additional point.
    
    :param doc: The Revit document.
    :type doc: Autodesk.Revit.DB.Document
    :param bounding_box: The bounding box of the element.
    :type bounding_box: Autodesk.Revit.DB.BoundingBoxXYZ
    :param additional_point: The additional point to draw lines to.
    :type additional_point: Autodesk.Revit.DB.XYZ
    :param view: The view in which to draw the lines.
    :type view: Autodesk.Revit.DB.View
    :param transaction_manager: The transaction manager to use.
    :type transaction_manager: function
    
    :return: Result object containing the status of the operation.
    :rtype: Result
    """

    return_value = Result()

    # get the corners of the bounding box
    min_x = bounding_box.Min.X
    min_y = bounding_box.Min.Y
    max_x = bounding_box.Max.X
    max_y = bounding_box.Max.Y

    # create a list of points for the lines
    points_bbox = [
        XYZ(min_x, min_y, 0),
        XYZ(max_x, min_y, 0),
        XYZ(max_x, max_y, 0),
        XYZ(min_x, max_y, 0),
        XYZ(min_x, min_y, 0),
    ]
    
    # flatten the additional point to 2D
    additional_point = XYZ(additional_point.X, additional_point.Y, 0)

    def action():
        action_return_value = Result()
        try:
            # draw the lines on the view
            for i in range(len(points_bbox) - 1):
                create_line_result = draw_line_between_points(doc, view, start_point=points_bbox[i], end_point=points_bbox[i + 1],  transaction_manager=None)
                action_return_value.update(create_line_result)
            # draw lines from each corner of the bounding box to the additional point
            for point in points_bbox[:-1]:
                create_line_result = draw_line_between_points(doc, view, start_point=point, end_point=additional_point,  transaction_manager=None)
                action_return_value.update(create_line_result)
        except Exception as e:
            action_return_value.update_sep(False, "Failed to draw lines: {}".format(e))

        return action_return_value

    # use the provided transaction manager to draw the bounding box
    transaction = Transaction(doc, "Drawing bounding box")
    return_value = transaction_manager(transaction, action)

    return return_value